from types import SimpleNamespace
import os
import numpy as np
import torch
from tqdm import tqdm

from .. import log
from ..utils import add_group_attr, dump_to_hdf5


class SolverBase:

    def __init__(self, wf=None, sampler=None,
                 optimizer=None, scheduler=None,
                 output=None, rank=0):
        """Base Class for QMC solver

        Args:
            wf (qmctorch.WaveFunction, optional): wave function. Defaults to None.
            sampler (qmctorch.sampler, optional): Sampler. Defaults to None.
            optimizer (torch.optim, optional): optimizer. Defaults to None.
            scheduler (torch.optim, optional): scheduler. Defaults to None.
            output (str, optional): hdf5 filename. Defaults to None.
            rank (int, optional): rank of he process. Defaults to 0.
        """

        self.wf = wf
        self.sampler = sampler
        self.opt = optimizer
        self.scheduler = scheduler
        self.cuda = False
        self.device = torch.device('cpu')

        # member defined in the child and or method
        self.dataloader = None
        self.loss = None
        self.obs_dict = None

        # if pos are needed for the optimizer (obsolete ?)
        if self.opt is not None and 'lpos_needed' not in self.opt.__dict__.keys():
            self.opt.lpos_needed = False

        # distributed model
        self.save_model = 'model.pth'

        # handles GPU availability
        if self.wf.cuda:
            self.device = torch.device('cuda')
            self.sampler.cuda = True
            self.sampler.walkers.cuda = True
        else:
            self.device = torch.device('cpu')

        self.hdf5file = output
        if output is None:
            basename = os.path.basename(
                self.wf.mol.hdf5file).split('.')[0]
            self.hdf5file = basename + '_QMCTorch.hdf5'

        if rank == 0:
            dump_to_hdf5(self, self.hdf5file)

        self.log_data()

    def configure_resampling(self, mode='update', resample_every=1, nstep_update=25):
        """Configure the resampling

        Args:
            mode (str, optional): method to resample : 'full', 'update', 'never'
                                  Defaults to 'update'.
            resample_every (int, optional): Number of optimization steps between resampling
                                 Defaults to 1.
            nstep_update (int, optional): Number of MC steps in update mode.
                                          Defaults to 25.
        """

        self.resampling_options = SimpleNamespace()
        valid_mode = ['never', 'full', 'update']
        if mode not in valid_mode:
            raise ValueError(
                mode, 'not a valid update method : ', valid_mode)

        self.resampling_options.mode = mode
        self.resampling_options.resample_every = resample_every
        self.resampling_options.nstep_update = nstep_update

    def track_observable(self, obs_name):
        """define the observalbe we want to track

        Args:
            obs_name (list): list of str defining the observalbe.
                             Each str must correspond to a WaveFuncion method
        """

        # make sure it's a list
        if not isinstance(obs_name, list):
            obs_name = list(obs_name)

        # sanity check
        valid_obs_name = ['energy', 'local_energy',
                          'geometry', 'parameters', 'gradients']
        for name in obs_name:
            if name in valid_obs_name:
                continue
            elif hasattr(self.wf, name):
                continue
            else:
                log.info(
                    '   Error : Observable %s not recognized' % name)
                log.info('         : Possible observable')
                for n in valid_obs_name:
                    log.info('         :  - %s' % n)
                log.info(
                    '         :  - or any method of the wave function')
                raise ValueError('Observable not recognized')

        # reset the Namesapce
        self.observable = SimpleNamespace()

        # add the energy of the sytem
        if 'energy' not in obs_name:
            obs_name += ['energy']

        # add the geometry of the system
        if 'geometry' not in obs_name:
            obs_name += ['geometry']

        for k in obs_name:

            if k == 'parameters':
                for key, p in zip(self.wf.state_dict().keys(),
                                  self.wf.parameters()):
                    if p.requires_grad:
                        self.observable.__setattr__(key, [])

            elif k == 'gradients':
                for key, p in zip(self.wf.state_dict().keys(),
                                  self.wf.parameters()):
                    if p.requires_grad:
                        self.observable.__setattr__(key+'.grad', [])

            else:
                self.observable.__setattr__(k, [])

        self.observable.models = SimpleNamespace()

    def store_observable(self, pos, local_energy=None, ibatch=None, **kwargs):
        """store observale in the dictionary

        Args:
            obs_dict (dict): dictionary of the observalbe
            pos (torch.tensor): positions of th walkers
            local_energy (torch.tensor, optional): precomputed values of the local
                                           energy. Defaults to None
            ibatch (int): index of the current batch. Defaults to None
        """

        if self.wf.cuda and pos.device.type == 'cpu':
            pos = pos.to(self.device)

        for obs in self.observable.__dict__.keys():

            # store the energy
            if obs == 'energy' and local_energy is not None:
                data = local_energy.cpu().detach().numpy()
                if (ibatch is None) or (ibatch == 0):
                    self.observable.energy.append(np.mean(data))
                else:
                    self.observable.energy[-1] *= ibatch/(ibatch+1)
                    self.observable.energy[-1] += np.mean(
                        data)/(ibatch+1)

            # store local energy
            elif obs == 'local_energy' and local_energy is not None:
                data = local_energy.cpu().detach().numpy()
                if (ibatch is None) or (ibatch == 0):
                    self.observable.local_energy.append(data)
                else:
                    self.observable.local_energy[-1] = np.append(
                        self.observable.local_energy[-1], data)

            # store variational parameter
            elif obs in self.wf.state_dict():

                p = self.wf.state_dict()[obs].clone()
                self.observable.__getattribute__(
                    obs).append(p.data.cpu().detach().numpy())

                if obs+'.grad' in self.observable.__dict__.keys():
                    if p.grad is not None:
                        self.observable.__getattribute__(obs +
                                                         '.grad').append(p.grad.cpu().numpy())
                    else:
                        self.observable.__getattribute__(obs +
                                                         '.grad').append(torch.zeros_like(p.data).cpu().numpy())

            # store any other defined method
            elif hasattr(self.wf, obs):
                func = self.wf.__getattribute__(obs)
                data = func(pos)
                if isinstance(data, torch.Tensor):
                    data = data.cpu().detach().numpy()
                if isinstance(data, list):
                    data = np.array(data)
                if (ibatch is None) or (ibatch == 0):
                    self.observable.__getattribute__(
                        obs).append(data)
                else:
                    self.observable.__getattribute__(
                        obs)[-1] = np.append(self.observable.__getattribute__(
                            obs)[-1], data)

    def print_observable(self, cumulative_loss, verbose=False):
        """Print the observalbe to csreen

        Args:
            cumulative_loss (float): current loss value
            verbose (bool, optional): print all the observables. Defaults to False
        """

        for k in self.observable.__dict__.keys():

            if k == 'local_energy':

                eloc = self.observable.local_energy[-1]
                e = np.mean(eloc)
                v = np.var(eloc)
                err = np.sqrt(v / len(eloc))
                log.options(style='percent').info(
                    '  energy   : %f +/- %f' % (e, err))
                log.options(style='percent').info(
                    '  variance : %f' % np.sqrt(v))

            elif verbose:
                log.options(style='percent').info(
                    k + ' : ', self.observable.__getattribute__(k)[-1])
                log.options(style='percent').info(
                    'loss %f' % (cumulative_loss))

    def resample(self, n, pos):
        """Resample the wave function

        Args:
            n (int): current epoch value
            pos (torch.tensor): positions of the walkers

        Returns:
            (torch.tensor): new positions of the walkers
        """

        if self.resampling_options.mode != 'never':

            # resample the data
            if (n % self.resampling_options.resample_every == 0):

                # make a copy of the pos if we update
                if self.resampling_options.mode == 'update':
                    pos = pos.clone().detach().to(self.device)

                # start from scratch otherwise
                else:
                    pos = None

                # sample and update the dataset
                pos = self.sampler(
                    self.wf.pdf, pos=pos, with_tqdm=False)
                self.dataloader.dataset.data = pos

            # update the weight of the loss if needed
            if self.loss.use_weight:
                self.loss.weight['psi0'] = None

        return pos

    def single_point(self, with_tqdm=True, hdf5_group='single_point'):
        """Performs a single point calculatin

        Args:
            with_tqdm (bool, optional): use tqdm for samplig. Defaults to True.
            hdf5_group (str, optional): hdf5 group where to store the data.
                                        Defaults to 'single_point'.

        Returns:
            SimpleNamespace: contains the local energy, positions, ...
        """

        log.info('')
        log.info('  Single Point Calculation : {nw} walkers | {ns} steps',
                 nw=self.sampler.nwalkers, ns=self.sampler.nstep)

        # check if we have to compute and store the grads
        grad_mode = torch.no_grad()
        if self.wf.kinetic == 'auto':
            grad_mode = torch.enable_grad()

        with grad_mode:

            #  get the position and put to gpu if necessary
            pos = self.sampler(self.wf.pdf, with_tqdm=with_tqdm)
            if self.wf.cuda and pos.device.type == 'cpu':
                pos = pos.to(self.device)

            # compute energy/variance/error
            el = self.wf.local_energy(pos)
            e, s, err = torch.mean(el), torch.var(
                el), self.wf.sampling_error(el)

            # print data
            log.options(style='percent').info(
                '  Energy   : %f +/- %f' % (e.detach().item(), err.detach().item()))
            log.options(style='percent').info(
                '  Variance : %f' % s.detach().item())

            # dump data to hdf5
            obs = SimpleNamespace(
                pos=pos,
                local_energy=el,
                energy=e,
                variance=s,
                error=err
            )
            dump_to_hdf5(obs,
                         self.hdf5file,
                         root_name=hdf5_group)
            add_group_attr(self.hdf5file, hdf5_group,
                           {'type': 'single_point'})

        return obs

    def save_checkpoint(self, epoch, loss):
        """save the model and optimizer state

        Args:
            epoch (int): epoch
            loss (float): current value of the loss
            filename (str): name to save the file
        """
        filename = 'checkpoint_epoch%d.pth' % epoch
        torch.save({
            'epoch': epoch,
            'model_state_dict': self.wf.state_dict(),
            'optimzier_state_dict': self.opt.state_dict(),
            'loss': loss
        }, filename)

    def load_checkpoint(self, filename):
        """load a model/optmizer

        Args:
            filename (str): filename

        Returns:
            tuple : epoch number and loss
        """
        data = torch.load(filename)
        self.wf.load_state_dict(data['model_state_dict'])
        self.opt.load_state_dict(data['optimzier_state_dict'])
        epoch = data['epoch']
        loss = data['loss']
        return epoch, loss

    def _append_observable(self, key, data):
        """Append a new data point to observable key.

        Arguments:
            key {str} -- name of the observable
            data {} -- data
        """

        if key not in self.obs_dict.keys():
            self.obs_dict[key] = []
        self.obs_dict[key].append(data)

    def sampling_traj(self, pos=None, with_tqdm=True, hdf5_group='sampling_trajectory'):
        """Compute the local energy along a sampling trajectory

        Args:
            pos (torch.tensor): positions of the walkers along the trajectory
            hdf5_group (str, optional): name of the group where to store the data.
                                        Defaults to 'sampling_trajecory'
        Returns:
            SimpleNamespace : contains energy/positions/
        """
        log.info('')
        log.info('  Sampling trajectory')

        if pos is None:
            pos = self.sampler(self.wf.pdf, with_tqdm=with_tqdm)

        ndim = pos.shape[-1]
        p = pos.view(-1, self.sampler.nwalkers, ndim)
        el = []
        rng = tqdm(p, desc='INFO:QMCTorch|  Energy  ',
                   disable=not with_tqdm)
        for ip in rng:
            el.append(self.wf.local_energy(ip).cpu().detach().numpy())

        el = np.array(el).squeeze(-1)
        obs = SimpleNamespace(local_energy=np.array(el), pos=pos)
        dump_to_hdf5(obs,
                     self.hdf5file, hdf5_group)

        add_group_attr(self.hdf5file, hdf5_group,
                       {'type': 'sampling_traj'})
        return obs

    def print_parameters(self, grad=False):
        """print parameter values

        Args:
            grad (bool, optional): also print the gradient. Defaults to False.
        """
        for p in self.wf.parameters():
            if p.requires_grad:
                if grad:
                    print(p.grad)
                else:
                    print(p)

    def optimization_step(self, lpos):
        """Performs one optimization step

        Arguments:
            lpos {torch.tensor} -- positions of the walkers
        """

        if self.opt.lpos_needed:
            self.opt.step(lpos)
        else:
            self.opt.step()

    def save_traj(self, fname, obs):
        """Save trajectory of geo_opt

        Args:
            fname (str): file name
        """
        f = open(fname, 'w')
        xyz = obs.geometry
        natom = len(xyz[0])
        nm2bohr = 1.88973
        for snap in xyz:
            f.write('%d \n\n' % natom)
            for i, pos in enumerate(snap):
                at = self.wf.atoms[i]
                f.write('%s % 7.5f % 7.5f %7.5f\n' % (at[0],
                                                      pos[0] /
                                                      nm2bohr,
                                                      pos[1] /
                                                      nm2bohr,
                                                      pos[2] / nm2bohr))
            f.write('\n')
        f.close()

    def run(self, nepoch, batchsize=None, loss='variance'):
        raise NotImplementedError()

    def log_data(self):
        """Log basic information about the sampler."""

        log.info('')
        log.info(' QMC Solver ')

        if self.wf is not None:
            log.info(
                '  WaveFunction        : {0}', self.wf.__class__.__name__)
            for x in self.wf.__repr__().split('\n'):
                log.debug('   ' + x)

        if self.sampler is not None:
            log.info(
                '  Sampler             : {0}', self.sampler.__class__.__name__)
            for x in self.sampler.__repr__().split('\n'):
                log.debug('   ' + x)

        if self.opt is not None:
            log.info(
                '  Optimizer           : {0}', self.opt.__class__.__name__)
            for x in self.opt.__repr__().split('\n'):
                log.debug('   ' + x)

        if self.scheduler is not None:
            log.info(
                '  Scheduler           : {0}', self.scheduler.__class__.__name__)
            for x in self.scheduler.__repr__().split('\n'):
                log.debug('   ' + x)



import torch
from ...utils import fast_power


def radial_slater(R, bas_n, bas_exp, xyz=None,
                  derivative=0, sum_grad=True, sum_hess=True):
    """Compute the radial part of STOs (or its derivative).

    .. math:
        sto = r^n exp(-\alpha |r|)

    Args:
        R (torch.tensor): distance between each electron and each atom
        bas_n (torch.tensor): principal quantum number
        bas_exp (torch.tensor): exponents of the exponential

    Keyword Arguments:
        xyz (torch.tensor): positions of the electrons
                            (needed for derivative) (default: {None})
        derivative (int): degree of the derivative (default: {0})
                          0 : value of the function
                          1 : first derivative
                          2 : pure second derivative
                          3 : mixed second derivative
        sum_grad (bool): return the sum_grad, i.e the sum of the gradients
                           (default: {True})
        sum_hess (bool): return the sum_hess, i.e the sum of the diag hessian
                           (default: {False})
        mixed_hess (bool): return the full hessian for each electron
                            i.e. dxdy dxdz dydz ... mixed derivatives
                           (default: {False})

    Returns:
        torch.tensor: values of each orbital radial part at each position
    """

    if not isinstance(derivative, list):
        derivative = [derivative]

    def _kernel():
        """Return the kernel."""
        return rn * er

    def _first_derivative_kernel():
        """Return the first derivative."""
        if sum_grad:
            nabla_rn_sum = nabla_rn.sum(3)
            nabla_er_sum = nabla_er.sum(3)
            return nabla_rn_sum * er + rn * nabla_er_sum
        else:
            return nabla_rn * \
                er.unsqueeze(-1) + rn.unsqueeze(-1) * nabla_er

    def _second_derivative_kernel():
        """Return the pure second derivative i.e. d^2/dx^2 """
        if sum_hess:

            lap_rn = nRnm2 * (bas_n + 1)
            lap_er = bexp_er * (bas_exp - 2. / R)

            return lap_rn * er + 2 * \
                (nabla_rn * nabla_er).sum(3) + rn * lap_er
        else:

            xyz2 = xyz*xyz
            xyz2 = xyz2 / xyz2.sum(-1, keepdim=True)

            lap_rn = nRnm2.unsqueeze(-1) * \
                (1. + (bas_n-2).unsqueeze(-1) * xyz2)

            lap_er = bexp_er.unsqueeze(-1) * \
                (bas_exp.unsqueeze(-1) * xyz2 +
                    (-1 + xyz2)/R.unsqueeze(-1))

            return lap_rn * er.unsqueeze(-1) + 2 * \
                (nabla_rn * nabla_er) + rn.unsqueeze(-1) * lap_er

    def _mixed_second_derivative_kernel():
        """Returns the mixed second derivative i.e. d^2/dxdy.
        where x and y are coordinate of the same electron."""

        mix_prod = xyz[..., [[0, 1], [0, 2], [1, 2]]].prod(-1)
        nRnm4 = nRnm2 / (xyz*xyz).sum(-1)

        lap_rn = ((bas_n-2) * nRnm4).unsqueeze(-1) * mix_prod

        lap_er = (bexp_er/(xyz*xyz).sum(-1)).unsqueeze(-1) * mix_prod * (
            bas_exp.unsqueeze(-1) + 1./R.unsqueeze(-1))

        return lap_rn * er.unsqueeze(-1) \
            + (nabla_rn[..., [[0, 1], [0, 2], [1, 2]]] * nabla_er[..., [[1, 0], [2, 0], [2, 1]]]).sum(-1) \
            + rn.unsqueeze(-1) * lap_er

    # computes the basic quantities
    rn = fast_power(R, bas_n)
    er = torch.exp(-bas_exp * R)

    # computes the grad
    if any(x in derivative for x in [1, 2, 3]):
        Rnm2 = R**(bas_n - 2)
        nRnm2 = bas_n * Rnm2
        bexp_er = bas_exp * er
        nabla_rn = (nRnm2).unsqueeze(-1) * xyz
        nabla_er = -(bexp_er).unsqueeze(-1) * \
            xyz / R.unsqueeze(-1)

    return return_required_data(derivative, _kernel,
                                _first_derivative_kernel,
                                _second_derivative_kernel,
                                _mixed_second_derivative_kernel)


def radial_gaussian(R, bas_n, bas_exp, xyz=None, derivative=[0],
                    sum_grad=True, sum_hess=True):
    """Compute the radial part of GTOs (or its derivative).

    .. math:
        gto = r ^ n exp(-\alpha r ^ 2)

    Args:
        R(torch.tensor): distance between each electron and each atom
        bas_n(torch.tensor): principal quantum number
        bas_exp(torch.tensor): exponents of the exponential

    Keyword Arguments:
        xyz(torch.tensor): positions of the electrons
                            (needed for derivative)(default: {None})
        derivative(int): degree of the derivative(default: {0})
        sum_grad(bool): return the sum_grad, i.e the sum of the gradients
                           (default: {True})

    Returns:
        torch.tensor: values of each orbital radial part at each position
    """

    if not isinstance(derivative, list):
        derivative = [derivative]

    def _kernel():
        return rn * er

    def _first_derivative_kernel():

        if sum_grad:
            nabla_rn_sum = nabla_rn.sum(3)
            nabla_er_sum = nabla_er.sum(3)
            return nabla_rn_sum * er + rn * nabla_er_sum
        else:
            return nabla_rn * \
                er.unsqueeze(-1) + rn.unsqueeze(-1) * nabla_er

    def _second_derivative_kernel():

        if sum_hess:
            lap_rn = nRnm2 * (bas_n + 1)
            lap_er = bas_exp * er * (4*bas_exp*R2 - 6)

            return lap_rn * er + 2 * \
                (nabla_rn * nabla_er).sum(3) + rn * lap_er

        else:
            xyz2 = xyz*xyz

            lap_er = (bas_exp * er).unsqueeze(-1) * \
                (4*bas_exp.unsqueeze(-1)*xyz2-2)

            xyz2 = xyz2 / xyz2.sum(-1, keepdim=True)

            lap_rn = nRnm2.unsqueeze(-1) * \
                (1. + (bas_n-2).unsqueeze(-1) * xyz2)

            return lap_rn * er.unsqueeze(-1) + 2 * \
                (nabla_rn * nabla_er) + rn.unsqueeze(-1) * lap_er

    def _mixed_second_derivative_kernel():
        """Returns the mixed second derivative i.e. d^2/dxdy.
        where x and y are coordinate of the same electron."""

        mix_prod = xyz[..., [[0, 1], [0, 2], [1, 2]]].prod(-1)
        nRnm4 = nRnm2 / (xyz*xyz).sum(-1)

        lap_rn = ((bas_n-2) * nRnm4).unsqueeze(-1) * mix_prod

        lap_er = 4 * (bexp_er * bas_exp).unsqueeze(-1) * mix_prod

        return lap_rn * er.unsqueeze(-1) \
            + (nabla_rn[..., [[0, 1], [0, 2], [1, 2]]] * nabla_er[..., [[1, 0], [2, 0], [2, 1]]]).sum(-1) \
            + rn.unsqueeze(-1) * lap_er

    # computes the basic  quantities
    R2 = R*R
    rn = fast_power(R, bas_n)
    er = torch.exp(-bas_exp * R2)

    # computes the grads
    if any(x in derivative for x in [1, 2, 3]):

        Rnm2 = R**(bas_n - 2)
        nRnm2 = bas_n * Rnm2
        bexp_er = bas_exp * er

        nabla_rn = (nRnm2).unsqueeze(-1) * xyz
        nabla_er = -2 * (bexp_er).unsqueeze(-1) * xyz

    return return_required_data(derivative, _kernel,
                                _first_derivative_kernel,
                                _second_derivative_kernel,
                                _mixed_second_derivative_kernel)


def radial_gaussian_pure(R, bas_n, bas_exp, xyz=None, derivative=[0],
                         sum_grad=True, sum_hess=True):
    """Compute the radial part of GTOs (or its derivative).

    .. math:
        gto = exp(-\alpha r ^ 2)

    Args:
        R(torch.tensor): distance between each electron and each atom
        bas_n(torch.tensor): principal quantum number
        bas_exp(torch.tensor): exponents of the exponential

    Keyword Arguments:
        xyz(torch.tensor): positions of the electrons
                            (needed for derivative)(default: {None})
        derivative(int): degree of the derivative(default: {0})
        sum_grad(bool): return the sum_grad, i.e the sum of the gradients
                           (default: {True})
        sum_hess(bool): return the sum_hess, i.e the sum of the lapacian
                           (default: {True})

    Returns:
        torch.tensor: values of each orbital radial part at each position
    """

    if not isinstance(derivative, list):
        derivative = [derivative]

    def _kernel():
        return er

    def _first_derivative_kernel():
        if sum_grad:
            return nabla_er.sum(3)
        else:
            return nabla_er

    def _second_derivative_kernel():
        if sum_hess:
            lap_er = bas_exp * er * (4*bas_exp*R2 - 6)
            return lap_er
        else:
            xyz2 = xyz*xyz
            lap_er = (bas_exp * er).unsqueeze(-1) * \
                (4*bas_exp.unsqueeze(-1)*xyz2-2)
            return lap_er

    def _mixed_second_derivative_kernel():
        """Returns the mixed second derivative i.e. d^2/dxdy.
        where x and y are coordinate of the same electron."""

        mix_prod = xyz[..., [[0, 1], [0, 2], [1, 2]]].prod(-1)
        lap_er = 4 * (bexp_er * bas_exp).unsqueeze(-1) * mix_prod

        return lap_er

    # computes the basic  quantities
    R2 = R*R
    er = torch.exp(-bas_exp * R2)

    # computes the grads
    if any(x in derivative for x in [1, 2, 3]):

        bexp_er = bas_exp * er
        nabla_er = -2 * (bexp_er).unsqueeze(-1) * xyz

    return return_required_data(derivative, _kernel,
                                _first_derivative_kernel,
                                _second_derivative_kernel,
                                _mixed_second_derivative_kernel)


def radial_slater_pure(R, bas_n, bas_exp, xyz=None, derivative=0,
                       sum_grad=True, sum_hess=True):
    """Compute the radial part of STOs (or its derivative).

    .. math:
        sto = exp(-\alpha | r |)

    Args:
        R(torch.tensor): distance between each electron and each atom
        bas_n(torch.tensor): principal quantum number
        bas_exp(torch.tensor): exponents of the exponential

    Keyword Arguments:
        xyz(torch.tensor): positions of the electrons
                            (needed for derivative)(default: {None})
        derivative(int): degree of the derivative(default: {0})
        sum_grad(bool): return the sum_grad, i.e the sum of the gradients
                           (default: {True})
        sum_hess(bool): return the sum_hess, i.e the sum of the laplacian
                           (default: {True})

    Returns:
        torch.tensor: values of each orbital radial part at each position
    """

    if not isinstance(derivative, list):
        derivative = [derivative]

    def _kernel():
        return er

    def _first_derivative_kernel():
        if sum_grad:
            return nabla_er.sum(3)
        else:
            return nabla_er

    def _second_derivative_kernel():

        if sum_hess:
            return bexp_er * (bas_exp - 2. / R)

        else:
            xyz2 = xyz*xyz / (R*R).unsqueeze(-1)
            lap_er = bexp_er.unsqueeze(-1) * \
                (bas_exp.unsqueeze(-1) * xyz2 - (1-xyz2)/R.unsqueeze(-1))
            return lap_er

    def _mixed_second_derivative_kernel():
        """Returns the mixed second derivative i.e. d^2/dxdy.
        where x and y are coordinate of the same electron."""

        mix_prod = xyz[..., [[0, 1], [0, 2], [1, 2]]].prod(-1)

        lap_er = (bexp_er/(xyz*xyz).sum(-1)).unsqueeze(-1) * mix_prod * (
            bas_exp.unsqueeze(-1) + 1./R.unsqueeze(-1))

        return lap_er

    # computes the basic gradients
    er = torch.exp(-bas_exp * R)

    # computes the grad
    if any(x in derivative for x in [1, 2, 3]):
        bexp_er = bas_exp * er
        nabla_er = -(bexp_er).unsqueeze(-1) * \
            xyz / R.unsqueeze(-1)

    return return_required_data(derivative, _kernel,
                                _first_derivative_kernel,
                                _second_derivative_kernel,
                                _mixed_second_derivative_kernel)


def return_required_data(derivative, _kernel,
                         _first_derivative_kernel,
                         _second_derivative_kernel,
                         _mixed_second_derivative_kernel):
    """Returns the data contained in derivative

    Args:
        derivative(list): list of the derivatives required
        _kernel(callable): kernel of the values
        _first_derivative_kernel(callable): kernel for 1st der
        _second_derivative_kernel(callable): kernel for 2nd der

    Returns:
        list: values of the different der requried
    """

    # prepare the output/kernel
    output = []
    fns = [_kernel,
           _first_derivative_kernel,
           _second_derivative_kernel,
           _mixed_second_derivative_kernel]

    # compute the requested functions
    for d in derivative:
        output.append(fns[d]())

    if len(derivative) == 1:
        return output[0]
    else:
        return output



import torch


class OrbitalConfigurations:

    def __init__(self, mol):
        # self.mol = mol
        self.nup = mol.nup
        self.ndown = mol.ndown
        self.nelec = self.nup + self.ndown
        self.norb = mol.basis.nmo

    def get_configs(self, configs):
        """Get the configuratio in the CI expansion

        Args:
            configs (str): name of the configs we want
            mol (mol object): molecule object

        Returns:
            tuple(torch.LongTensor,torch.LongTensor): the spin up/spin down
            electronic confs
        """

        if isinstance(configs, str):
            configs = configs.lower()

        if isinstance(configs, torch.Tensor):
            return configs

        elif configs == 'ground_state':
            return self._get_ground_state_config()

        elif configs.startswith('cas('):
            nelec, norb = eval(configs.lstrip("cas"))
            self.sanity_check(nelec, norb)
            nocc, nvirt = self._get_orb_number(nelec, norb)
            return self._get_cas_config(nocc, nvirt, nelec)

        elif configs.startswith('single('):
            nelec, norb = eval(configs.lstrip("single"))
            self.sanity_check(nelec, norb)
            nocc, nvirt = self._get_orb_number(nelec, norb)
            return self._get_single_config(nocc, nvirt)

        elif configs.startswith('single_double('):
            nelec, norb = eval(configs.lstrip("single_double"))
            self.sanity_check(nelec, norb)
            nocc, nvirt = self._get_orb_number(nelec, norb)
            return self._get_single_double_config(nocc, nvirt)

        else:
            print(configs, " not recognized as valid configuration")
            print('Options are : ground_state')
            print('              single(nelec,norb)')
            print('              single_double(nelec,norb)')
            print('              cas(nelec,norb)')
            raise ValueError("Config error")

    def sanity_check(self, nelec, norb):
        """Check if the number of elec/orb is consistent with the
           properties of the molecule

        Args:
            nelec (int): required number of electrons in config
            norb (int): required number of orb in config

        """
        if nelec > self.nelec:
            raise ValueError(
                'required number of electron in config too large')

        if norb > self.norb:
            raise ValueError(
                'required number of orbitals in config too large')

    def _get_ground_state_config(self):
        """Return only the ground state configuration

        Args:
            mol (mol): mol object

        Returns:
            tuple(torch.LongTensor,torch.LongTensor): the spin up/spin down
            electronic confs
        """
        _gs_up = list(range(self.nup))
        _gs_down = list(range(self.ndown))
        cup, cdown = [_gs_up], [_gs_down]
        return (torch.LongTensor(cup), torch.LongTensor(cdown))

    def _get_single_config(self, nocc, nvirt):
        """Get the confs of the singlet conformations

        Args:
            mol (mol): mol object
            nocc (int): number of occupied orbitals in the active space
            nvirt (int): number of virtual orbitals in the active space
        """

        _gs_up = list(range(self.nup))
        _gs_down = list(range(self.ndown))
        cup, cdown = [_gs_up], [_gs_down]

        for iocc in range(
                self.nup - 1, self.nup - 1 - nocc, -1):
            for ivirt in range(self.nup, self.nup + nvirt, 1):
                _xt = self._create_excitation(
                    _gs_up.copy(), iocc, ivirt)
                cup, cdown = self._append_excitations(
                    cup, cdown, _xt, _gs_down)

                _xt = self._create_excitation(
                    _gs_down.copy(), iocc, ivirt)
                cup, cdown = self._append_excitations(
                    cup, cdown, _gs_up, _xt)

        return (torch.LongTensor(cup), torch.LongTensor(cdown))

    def _get_single_double_config(self, nocc, nvirt):
        """Get the confs of the single + double

        Args:
            nelec (int): number of electrons in the active space
            norb (int) : number of orbitals in the active space
        """

        _gs_up = list(range(self.nup))
        _gs_down = list(range(self.ndown))
        cup, cdown = self._get_single_config(nocc, nvirt)
        cup = cup.tolist()
        cdown = cdown.tolist()

        idx_occ_up = list(
            range(self.nup - 1, self.nup - 1 - nocc, -1))
        idx_vrt_up = list(range(self.nup, self.nup + nvirt, 1))

        idx_occ_down = list(range(
            self.ndown - 1, self.ndown - 1 - nocc, -1))
        idx_vrt_down = list(range(self.ndown, self.ndown + nvirt, 1))

        # ground, single and double with 1 elec excited per spin
        for iocc_up in idx_occ_up:
            for ivirt_up in idx_vrt_up:

                for iocc_down in idx_occ_down:
                    for ivirt_down in idx_vrt_down:

                        _xt_up = self._create_excitation(
                            _gs_up.copy(), iocc_up, ivirt_up)
                        _xt_down = self._create_excitation(
                            _gs_down.copy(), iocc_down, ivirt_down)
                        cup, cdown = self._append_excitations(
                            cup, cdown, _xt_up, _xt_down)

        # double with 2elec excited on spin up
        for occ1, occ2 in torch.combinations(torch.as_tensor(idx_occ_up), r=2):
            for vrt1, vrt2 in torch.combinations(torch.as_tensor(idx_vrt_up), r=2):
                _xt_up = self._create_excitation(
                    _gs_up.copy(), occ1, vrt2)
                _xt_up = self._create_excitation(_xt_up, occ2, vrt1)
                cup, cdown = self._append_excitations(
                    cup, cdown, _xt_up, _gs_down)

        # double with 2elec excited per spin
        for occ1, occ2 in torch.combinations(torch.as_tensor(idx_occ_down), r=2):
            for vrt1, vrt2 in torch.combinations(torch.as_tensor(idx_vrt_down), r=2):

                _xt_down = self._create_excitation(
                    _gs_down.copy(), occ1, vrt2)
                _xt_down = self._create_excitation(
                    _xt_down, occ2, vrt1)
                cup, cdown = self._append_excitations(
                    cup, cdown, _gs_up, _xt_down)

        return (torch.LongTensor(cup), torch.LongTensor(cdown))

    def _get_cas_config(self, nocc, nvirt, nelec):
        """get confs of the CAS

        Args:
            nocc (int): number of occupied orbitals in the CAS
            nvirt ([type]): number of virt orbitals in the CAS
        """
        from itertools import combinations, product

        idx_low, idx_high = self.nup - nocc, self.nup + nvirt
        orb_index_up = range(idx_low, idx_high)
        idx_frz = list(range(idx_low))
        _cup = [idx_frz + list(l)
                for l in list(combinations(orb_index_up, nelec // 2))]

        idx_low, idx_high = self.nup - nocc - 1, self.nup + nvirt - 1

        _cdown = [
            idx_frz +
            list(l) for l in list(
                combinations(orb_index_up, nelec // 2))]

        confs = list(product(_cup, _cdown))
        cup, cdown = [], []

        for c in confs:
            cup.append(c[0])
            cdown.append(c[1])

        return (torch.LongTensor(cup), torch.LongTensor(cdown))

    def _get_orb_number(self, nelec, norb):
        """compute the number of occupied and virtual orbital
        __ PER SPIN __
        __ ONLY VALID For spin up/down ___
        Args:
            nelec (int): total number of elec in the CAS
            norb (int): total number of orb in the CAS

        Returns:
            [int,int]: number of occpuied/virtual orb per spi
        """

        nocc = nelec // 2
        nvirt = norb - nocc
        return nocc, nvirt

    def _create_excitation(self, conf, iocc, ivirt):
        return self._create_excitation_replace(conf, iocc, ivirt)

    @staticmethod
    def _create_excitation_ordered(conf, iocc, ivirt):
        """promote an electron from iocc to ivirt

        Args:
            conf (list): index of the occupied orbitals
            iocc (int): index of the occupied orbital
            ivirt (int): index of the virtual orbital

        Returns:
            list: new configuration by increasing order
                  e.g: 4->6 leads to : [0,1,2,3,5,6]
        Note:
            if that method is used to define the exciation index
            permutation must be accounted for when  computing
            the determinant as
            det(A[:,[0,1,2,3]]) = -det(A[:,[0,1,3,2]])
            see : ExcitationMask.get_index_unique_single()
                  in oribtal_projector.py
        """
        conf.pop(iocc)
        conf += [ivirt]
        return conf

    @staticmethod
    def _create_excitation_replace(conf, iocc, ivirt):
        """promote an electron from iocc to ivirt

        Args:
            conf (list): index of the occupied orbitals
            iocc (int): index of the occupied orbital
            ivirt (int): index of the virtual orbital

        Returns:
            list: new configuration not ordered
                e.g.: 4->6 leads tpo : [0,1,2,3,6,5]
        """
        conf[iocc] = ivirt
        return conf

    @staticmethod
    def _append_excitations(cup, cdown, new_cup, new_cdown):
        """Append new excitations

        Args:
            cup (list): configurations of spin up
            cdown (list): configurations of spin down
            new_cup (list): new spin up confs
            new_cdown (list): new spin down confs
        """

        cup.append(new_cup)
        cdown.append(new_cdown)
        return cup, cdown


def get_excitation(configs):
    """get the excitation data

    Args:
        configs (tuple): configuratin of the electrons

    Returns:
        exc_up, exc_down : index of the obitals in the excitaitons
                            [i,j],[l,m] : excitation i -> l, j -> l
    """
    exc_up, exc_down = [], []
    for ic, (cup, cdown) in enumerate(zip(configs[0], configs[1])):

        set_cup = set(tuple(cup.tolist()))
        set_cdown = set(tuple(cdown.tolist()))

        if ic == 0:
            set_gs_up = set_cup
            set_gs_down = set_cdown

        else:
            exc_up.append([list(set_gs_up.difference(set_cup)),
                           list(set_cup.difference(set_gs_up))])

            exc_down.append([list(set_gs_down.difference(set_cdown)),
                             list(set_cdown.difference(set_gs_down))])

    return (exc_up, exc_down)


def get_unique_excitation(configs):
    """get the unique excitation data

    Args:
        configs (tuple): configuratin of the electrons

    Returns:
        exc_up, exc_down : index of the obitals in the excitaitons
                            [i,j],[l,m] : excitation i -> l, j -> l
        index_up, index_down : index map for the unique exc
                                [0,0,...], [0,1,...] means that
                                1st : excitation is composed of unique_up[0]*unique_down[0]
                                2nd : excitation is composed of unique_up[0]*unique_down[1]
                                ....

    """
    uniq_exc_up, uniq_exc_down = [], []
    index_uniq_exc_up, index_uniq_exc_down = [], []
    for ic, (cup, cdown) in enumerate(zip(configs[0], configs[1])):

        set_cup = set(tuple(cup.tolist()))
        set_cdown = set(tuple(cdown.tolist()))

        if ic == 0:
            set_gs_up = set_cup
            set_gs_down = set_cdown

        exc_up = [list(set_gs_up.difference(set_cup)),
                  list(set_cup.difference(set_gs_up))]

        exc_down = [list(set_gs_down.difference(set_cdown)),
                    list(set_cdown.difference(set_gs_down))]

        if exc_up not in uniq_exc_up:
            uniq_exc_up.append(exc_up)

        if exc_down not in uniq_exc_down:
            uniq_exc_down.append(exc_down)

        index_uniq_exc_up.append(uniq_exc_up.index(exc_up))
        index_uniq_exc_down.append(
            uniq_exc_down.index(exc_down))

    return (uniq_exc_up, uniq_exc_down), (index_uniq_exc_up, index_uniq_exc_down)



import unittest

from qmctorch.sampler import Hamiltonian
from .test_sampler_base import TestSamplerBase


class TestHamiltonian(TestSamplerBase):

    def test_hmc(self):
        """Test HMC sampler."""
        sampler = Hamiltonian(
            nwalkers=10,
            nstep=20,
            step_size=0.1,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'))

        pos = sampler(self.wf.pdf)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import torch

from qmctorch.utils import set_torch_double_precision
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow


class TestSamplerBase(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        self.mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # orbital
        self.wf = SlaterJastrow(self.mol)



import unittest

from qmctorch.sampler.walkers import Walkers
from .test_sampler_base import TestSamplerBase


class TestWalkers(TestSamplerBase):

    def test_walkers_init(self):
        """Test different initialization methods of the walkers."""
        w1 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('center'))

        w2 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('uniform'))

        w3 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('normal'))

        w4 = Walkers(nwalkers=10,
                     nelec=self.mol.nelec, ndim=3,
                     init=self.mol.domain('atomic'))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Documentation about QMCTorch"""

from .__version__ import __version__

__author__ = "Nicolas Renaud"
__email__ = 'n.renaud@esciencecenter.nl'

import twiggy
import sys
twiggy.quick_setup(file=sys.stdout)
log = twiggy.log.name('QMCTorch')
log.min_level = twiggy.levels.INFO

log.info(r"  ____    __  ______________             _")
log.info(r" / __ \  /  |/  / ___/_  __/__  ________/ /  ")
log.info(r"/ /_/ / / /|_/ / /__  / / / _ \/ __/ __/ _ \ ")
log.info(r"\___\_\/_/  /_/\___/ /_/  \___/_/  \__/_//_/ ")



import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Hamiltonian, Metropolis
from qmctorch.solver import Solver
from qmctorch.utils import (plot_block, plot_blocking_energy,
                            plot_correlation_coefficient, plot_energy,
                            plot_integrated_autocorrelation_time,
                            plot_walkers_traj)
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow

__PLOT__ = True


class TestH2(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)

        # optimal parameters
        self.opt_r = 0.69  # the two h are at +0.69 and -0.69
        self.opt_sigma = 1.24

        # molecule
        self.mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='auto',
                                configs='single(2,2)')

        # sampler
        self.sampler = Metropolis(
            nwalkers=1000,
            nstep=2000,
            step_size=0.5,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        self.hmc_sampler = Hamiltonian(
            nwalkers=100,
            nstep=200,
            step_size=0.1,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'))

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

        # ground state energy
        self.ground_state_energy = -1.16

        # ground state pos
        self.ground_state_pos = 0.69

    def test1_single_point(self):

        self.solver.wf.ao.atom_coords[0, 2] = -self.ground_state_pos
        self.solver.wf.ao.atom_coords[1, 2] = self.ground_state_pos
        self.solver.sampler = self.sampler

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        # values on different arch
        expected_energy = [-1.1464850902557373,
                           -1.14937478612449]

        # values on different arch
        expected_variance = [0.9279592633247375,
                             0.7445300449383236]

        assert(np.any(np.isclose(e.data.item(), np.array(expected_energy))))
        assert(np.any(np.isclose(v.data.item(), np.array(expected_variance))))

    def test2_single_point_hmc(self):

        self.solver.wf.ao.atom_coords[0, 2] = -self.ground_state_pos
        self.solver.wf.ao.atom_coords[1, 2] = self.ground_state_pos
        self.solver.sampler = self.hmc_sampler

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        # values on different arch
        expected_energy = [-1.0877732038497925,
                           -1.088576]

        # values on different arch
        expected_variance = [0.14341972768306732,
                             0.163771]

        assert(np.any(np.isclose(e.data.item(), np.array(expected_energy))))
        assert(np.any(np.isclose(v.data.item(), np.array(expected_variance))))

    def test3_wf_opt(self):
        self.solver.sampler = self.sampler

        self.solver.configure(track=['local_energy', 'parameters'],
                              loss='energy', grad='auto')
        obs = self.solver.run(5)
        if __PLOT__:
            plot_energy(obs.local_energy, e0=-
                        1.1645, show_variance=True)

    def test4_geo_opt(self):

        self.solver.wf.ao.atom_coords[0,
                                      2].data = torch.as_tensor(-0.37)
        self.solver.wf.ao.atom_coords[1,
                                      2].data = torch.as_tensor(0.37)

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='auto')
        self.solver.geo_opt(5, nepoch_wf_init=10, nepoch_wf_update=5)

        # load the best model
        self.solver.wf.load(self.solver.hdf5file, 'geo_opt')
        self.solver.wf.eval()

        # sample and compute variables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        e = e.data.numpy()
        v = v.data.numpy()

        # it might be too much to assert with the ground state energy
        assert(e > 2 * self.ground_state_energy and e < 0.)
        assert(v > 0 and v < 2.)

    def test5_sampling_traj(self):
        self.solver.sampler = self.sampler

        self.solver.sampler.nstep = 100
        self.solver.sampler.ntherm = 0
        self.solver.sampler.ndecor = 1

        pos = self.solver.sampler(self.solver.wf.pdf)
        obs = self.solver.sampling_traj(pos)

        if __PLOT__:
            plot_walkers_traj(obs.local_energy)
            plot_block(obs.local_energy)

            plot_blocking_energy(obs.local_energy, block_size=10)
            plot_correlation_coefficient(obs.local_energy)
            plot_integrated_autocorrelation_time(obs.local_energy)


if __name__ == "__main__":
    # unittest.main()
    t = TestH2()
    t.setUp()
    # t.test2_single_point_hmc()
    # t.test1_single_point()
    t.test3_wf_opt()
    # t.test5_sampling_traj()



import os
import math
import numpy as np
from mendeleev import element
from types import SimpleNamespace
import h5py
from mpi4py import MPI
from .calculator import CalculatorADF, CalculatorPySCF, CalculatorADF2019

from ..utils import dump_to_hdf5, load_from_hdf5, bytes2str
from .. import log


class Molecule:

    def __init__(self, atom=None, calculator='pyscf',
                 scf='hf', basis='sto-3g', unit='bohr',
                 name=None, load=None, save_scf_file=False,
                 redo_scf=False, rank=0):
        """Create a molecule in QMCTorch

        Args:
            atom (str or None, optional): defines the atoms and their positions. Defaults to None.
                - At1 x y z; At2 x y z ... : Provide the atomic coordinate directly
                - <file>.xyz : provide the path to an .xyz file containing the atomic coordinates 
            calculator (str, optional): selet scf calculator. Defaults to 'adf'.
                - pyscf : PySCF calculator
                - adf : ADF2020+ calculator
                - adf2019 : ADF2019 calculatori
            scf (str, optional): select scf level of theory. Defaults to 'hf'.
                - hf : perform a Hatree-Fock calculation to obtain the molecular orbital coefficients
                - dft : perform a density functional theory using the local density approximation
            basis (str, optional): select the basis set. Defaults to 'dzp'.
            unit (str, optional): units of the coordinates; 'bohr' or 'angs'. Defaults to 'bohr'.
            name (str or None, optional): name of the molecule. Defaults to None.
            load (str or None, optional): path to a hdf5 file to load. Defaults to None.
            save_scf_file (bool, optional): save the scf file (when applicable) Defaults to False
            redo_scf (bool, optional): if true ignore existing hdf5 file and redo the scf calculation
            rank (int, optional): Rank of the process. Defaults to 0.

        Examples:
            >>> from qmctorch.scf import Molecule
            >>> mol = Molecule(atom='H 0 0 0; H 0 0 1', unit='angs',
            ...                calculator='adf', basis='dzp')
        """

        self.atom_coords = []
        self.atomic_nelec = []
        self.atomic_number = []
        self.atoms = []
        self.atoms_str = atom
        self.hdf5file = None
        self.max_angular = 2
        self.name = name
        self.natom = 0
        self.ndown = 0
        self.nelec = 0
        self.nup = 0
        self.unit = unit
        self.basis = SimpleNamespace()
        self.calculator_name = calculator
        self.basis_name = basis
        self.save_scf_file = save_scf_file
        self.scf_level = scf

        if rank == 0:
            log.info('')
            log.info(' SCF Calculation')

        # load an existing hdf5 file
        if load is not None:
            log.info('  Loading data from {file}', file=load)
            self._load_hdf5(load)
            self.hdf5file = load

        else:

            # extract the atom names/positions from
            # the atom kwargs
            self._process_atom_str()

            # name of the hdf5 file
            self.hdf5file = '_'.join(
                [self.name, calculator, basis]) + '.hdf5'

            if rank == 0:

                if self.unit not in ['angs', 'bohr']:
                    raise ValueError('unit should be angs or bohr')

                # force a redo of the sc calculation
                if os.path.isfile(self.hdf5file) and redo_scf:
                    log.info('  Removing {file} and redo SCF calculations',
                             file=self.hdf5file)
                    os.remove(self.hdf5file)

                # deals with existing files
                if os.path.isfile(self.hdf5file):
                    log.info('  Reusing scf results from {file}',
                             file=self.hdf5file)
                    self.basis = self._load_basis()

                # perform the scf calculation
                else:
                    log.info('  Running scf  calculation')

                    calc = {'adf2019': CalculatorADF2019,
                            'adf': CalculatorADF,
                            'pyscf': CalculatorPySCF}[calculator]

                    self.calculator = calc(self.atoms,
                                           self.atom_coords,
                                           basis,
                                           self.scf_level,
                                           self.unit,
                                           self.name,
                                           self.save_scf_file)

                    self.basis = self.calculator.run()
                    self.save_scf_file = self.calculator.savefile

                    dump_to_hdf5(self, self.hdf5file,
                                 root_name='molecule')

                    self._check_basis()
                    self.log_data()

            MPI.COMM_WORLD.barrier()
            if rank != 0:
                log.info(
                    '  Loading data from {file}', file=self.hdf5file)
                self._load_hdf5(self.hdf5file)

    def log_data(self):

        log.info('  Molecule name       : {0}', self.name)
        log.info('  Number of electrons : {0}', self.nelec)
        log.info(
            '  SCF calculator      : {0}', self.calculator_name)
        log.info('  Basis set           : {0}', self.basis_name)
        log.info(
            '  SCF                 : {0}', self.scf_level.upper())
        log.info('  Number of AOs       : {0}', self.basis.nao)
        log.info('  Number of MOs       : {0}', self.basis.nmo)
        log.info(
            '  SCF Energy          : {:.3f} Hartree'.format(self.get_total_energy()))

    def domain(self, method):
        """Returns information to initialize the walkers

        Args:
            method (str): 'center', all electron at the center of the system
                          'uniform', all electrons in a cube surrounding the molecule
                          'normal',  all electrons in a sphere surrounding the molecule
                          'atomic', electrons around the atoms

        Returns:
            dict: dictionary containing corresponding information

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> domain = mol.domain('atomic')
        """
        domain = dict()
        domain['method'] = method

        if method == 'center':
            domain['center'] = np.mean(self.atom_coords, 0)

        elif method == 'uniform':
            domain['min'] = np.min(self.atom_coords) - 0.5
            domain['max'] = np.max(self.atom_coords) + 0.5

        elif method == 'normal':
            domain['mean'] = np.mean(self.atom_coords, 0)
            domain['sigma'] = np.diag(
                np.std(self.atom_coords, 0) + 0.25)

        elif method == 'atomic':
            domain['atom_coords'] = self.atom_coords
            domain['atom_num'] = self.atomic_number
            domain['atom_nelec'] = self.atomic_nelec

        else:
            raise ValueError(
                'Method to initialize the walkers not recognized')

        return domain

    def _process_atom_str(self):
        """Process the atom description."""

        if self.atoms_str.endswith('.xyz'):
            if os.path.isfile(self.atoms_str):
                atoms = self._read_xyz_file()
            else:
                raise FileNotFoundError(
                    'File %s not  found' % self.atoms_str)
        else:
            atoms = self.atoms_str.split(';')

        self._get_atomic_properties(atoms)

    def _get_atomic_properties(self, atoms):
        """Generates the atomic propeties of the molecule

        Args:
            atoms (str): atoms given in input
        """

        # loop over all atoms
        for a in atoms:
            atom_data = a.split()
            self.atoms.append(atom_data[0])
            x, y, z = float(atom_data[1]), float(
                atom_data[2]), float(atom_data[3])

            conv2bohr = 1
            if self.unit == 'angs':
                conv2bohr = 1.88973
            self.atom_coords.append(
                [x * conv2bohr, y * conv2bohr, z * conv2bohr])

            self.atomic_number.append(
                element(atom_data[0]).atomic_number)
            self.atomic_nelec.append(element(atom_data[0]).electrons)
            self.nelec += element(atom_data[0]).electrons

        # size of the system
        self.natom = len(self.atoms)
        if self.nelec % 2 != 0:
            raise ValueError("Only equal spin up/down supported.")
        self.nup = math.ceil(self.nelec / 2)
        self.ndown = math.floor(self.nelec / 2)

        # name of the system
        if self.name is None:
            self.name = self._get_mol_name(self.atoms)
        self.atoms = np.array(self.atoms)

    def _read_xyz_file(self):
        """Process a xyz file containing the data

        Returns:
            list -- atoms and xyz position
        """
        with open(self.atoms_str, 'r') as f:
            data = f.readlines()

        natom = int(data[0])
        atoms = data[2:2+natom]
        self.atoms_str = ''
        for a in atoms[:-1]:
            self.atoms_str += a + '; '
        self.atoms_str += atoms[-1]
        return atoms

    @staticmethod
    def _get_mol_name(atoms):
        mol_name = ''
        unique_atoms = list(set(atoms))
        for ua in unique_atoms:
            mol_name += ua
            nat = atoms.count(ua)
            if nat > 1:
                mol_name += str(nat)
        return mol_name

    def _load_basis(self):
        """Get the basis information needed to compute the AO values."""

        h5 = h5py.File(self.hdf5file, 'r')
        basis_grp = h5['molecule']['basis']
        self.basis = SimpleNamespace()

        self.basis.radial_type = bytes2str(
            basis_grp['radial_type'][()])
        self.basis.harmonics_type = bytes2str(
            basis_grp['harmonics_type'][()])

        self.basis.nao = int(basis_grp['nao'][()])
        self.basis.nmo = int(basis_grp['nmo'][()])

        self.basis.nshells = basis_grp['nshells'][()]
        self.basis.nao_per_atom = basis_grp['nao_per_atom'][()]
        self.basis.index_ctr = basis_grp['index_ctr'][()]
        self.basis.nctr_per_ao = basis_grp['nctr_per_ao'][()]

        self.basis.bas_exp = basis_grp['bas_exp'][()]
        self.basis.bas_coeffs = basis_grp['bas_coeffs'][()]

        self.basis.atom_coords_internal = basis_grp['atom_coords_internal'][(
        )]

        self.basis.TotalEnergy = basis_grp['TotalEnergy'][()]
        self.basis.mos = basis_grp['mos'][()]

        if self.basis.harmonics_type == 'cart':
            self.basis.bas_kr = basis_grp['bas_kr'][()]
            self.basis.bas_kx = basis_grp['bas_kx'][()]
            self.basis.bas_ky = basis_grp['bas_ky'][()]
            self.basis.bas_kz = basis_grp['bas_kz'][()]

        elif self.basis.harmonics_type == 'sph':
            self.basis.bas_n = basis_grp['bas_n'][()]
            self.basis.bas_l = basis_grp['bas_l'][()]
            self.basis.bas_m = basis_grp['bas_m'][()]

        else:
            raise ValueError(
                'Harmonics type should be cart or sph \
                but %s was found in %s' % (self.basis.harmonics_type,
                                           self.hdf5file))

        h5.close()
        return self.basis

    def print_total_energy(self):
        """Print the SCF energy of the molecule.

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis='sz')
            >>> mol.print_total_energy()
        """
        e = self.get_total_energy()
        log.info('== SCF Energy : {e}', e=e)

    def get_total_energy(self):
        """Get the value of the total energy."""
        h5 = h5py.File(self.hdf5file, 'r')
        e = h5['molecule']['basis']['TotalEnergy'][()]
        h5.close()
        return e

    def _check_basis(self):
        """Check if the basis contains all the necessary fields."""

        names = ['bas_coeffs', 'bas_exp', 'nshells',
                 'atom_coords_internal', 'nao', 'nmo',
                 'index_ctr', 'mos', 'TotalEnergy']

        if self.basis.harmonics_type == 'cart':
            names += ['bas_kx', 'bas_ky', 'bas_kz', 'bas_kr']

        elif self.basis.harmonics_type == 'sph':
            names += ['bas_n', 'bas_l', 'bas_m']

        for n in names:
            if not hasattr(self.basis, n):
                raise ValueError(n, ' not in the basis namespace')

    def _load_hdf5(self, filename):
        """Load a molecule from hdf5

        Args:
            filename (str): path to the file to be loaded
        """

        # load the data
        load_from_hdf5(self, filename, 'molecule')

        # cast some of the important data type
        # should be done by the hdf5_utils in the future
        self.atoms = self.atoms.astype('U')
        self.basis.nao = int(self.basis.nao)
        self.basis.nmo = int(self.basis.nmo)

        cast_fn = {'nelec': int,
                   'nup': int,
                   'ndown': int,
                   'atoms': lambda x: x.astype('U'),
                   'atomic_nelec': lambda x: [int(i) for i in x]}

        for name, fn in cast_fn.items():
            self.__setattr__(name, fn(self.__getattribute__(name)))

        cast_fn = {'nao': int,
                   'nmo': int}

        for name, fn in cast_fn.items():
            self.basis.__setattr__(
                name, fn(self.basis.__getattribute__(name)))



import torch
from torch import nn

from .....utils import register_extra_attributes
from .jastrow_kernel_electron_nuclei_base import JastrowKernelElectronNucleiBase


class PadeJastrowKernel(JastrowKernelElectronNucleiBase):

    def __init__(self, nup, ndown, atomic_pos, cuda, w=1.):
        r"""Computes the Simple Pade-Jastrow factor

        .. math::
            J = \prod_{i<j} \exp(B_{ij}) \quad \quad \\text{with} \quad \quad
            B_{ij} = \\frac{w_0 r_{i,j}}{1 + w r_{i,j}}

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            atoms (torch.tensor): atomic positions of the atoms
            w (float, optional): Value of the variational parameter. Defaults to 1..
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__(nup, ndown, atomic_pos, cuda)

        self.weight = nn.Parameter(
            torch.as_tensor([w]), requires_grad=True).to(self.device)
        register_extra_attributes(self, ['weight'])

        self.static_weight = torch.as_tensor([1.]).to(self.device)
        self.requires_autograd = True

    def forward(self, r):
        """ Get the jastrow kernel.
        .. math::
            B_{ij} = \frac{b r_{i,j}}{1+b'r_{i,j}}

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: matrix of the jastrow kernels
                          Nbatch x Nelec x Nelec
        """
        return self.static_weight * r / (1.0 + self.weight * r)

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons

        .. math::

            d B_{ij} / d k_i =  d B_{ij} / d k_j  = - d B_{ji} / d k_i

            out_{k,i,j} = A1 + A2
            A1_{kij} = w0 \frac{dr_{ij}}{dk_i} / (1 + w r_{ij})
            A2_{kij} = - w0 w' r_{ij} \frac{dr_{ij}}{dk_i} / (1 + w r_{ij})^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        r_ = r.unsqueeze(1)
        denom = 1. / (1.0 + self.weight * r_)
        a = self.static_weight * dr * denom
        b = - self.static_weight * self.weight * r_ * dr * denom**2

        return (a + b)

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron

        .. math ::

            d^2 B_{ij} / d k_i^2 =  d^2 B_{ij} / d k_j^2 = d^2 B_{ji} / d k_i^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
            d2r (torch.tensor): matrix of the 2nd derivative of
                                the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of
                          the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        r_ = r.unsqueeze(1)
        denom = 1. / (1.0 + self.weight * r_)
        denom2 = denom**2
        dr_square = dr*dr

        a = self.static_weight * d2r * denom
        b = -2 * self.static_weight * self.weight * dr_square * denom2
        c = - self.static_weight * self.weight * r_ * d2r * denom2
        d = 2 * self.static_weight * self.weight**2 * r_ * dr_square * denom**3

        return a + b + c + d



import os
import unittest

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.autograd import Variable

from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow

from ...path_utils import PATH_TEST

__PLOT__ = False


def read_cubefile(fname):
    with open(fname, 'r') as f:
        data = f.readlines()
    vals = []
    for d in data[7:]:
        vals.append(float(d.split('\n')[0]))
    return vals


def get_pts(npts):
    x = torch.linspace(-1, 1, npts)
    xx, yy = torch.meshgrid((x, x))
    pts = torch.cat((xx.reshape(-1, 1), yy.reshape(-1, 1)), axis=1)
    pts = torch.cat((pts, torch.zeros(pts.shape[0], 1)), axis=1)
    return pts


def generate_cube_files(t21file):

    nao = create_ao_variable_in_t21(t21file)
    create_densf_input_file(t21file, nao)
    os.system('$ADFBIN/densf < densf_input')


def create_ao_variable_in_t21(t21file):

    from scm import plams
    with plams.kFFile(t21file) as kf:
        nao = kf.read('Basis', 'naos')
        for iao in range(nao):

            var = [0.] * nao
            var[iao] = 1.
            name = 'AO%d' % iao
            kf.write('Basis', name, var)

    return nao


def create_densf_input_file(t21name, nao):

    f = open('densf_input', 'w')
    f.write('INPUTFILE %s\n\nCUBOUTPUT C_AO_\n\n' % t21name)

    f.write('GRID \n')
    f.write(' -1 -1 0\n')
    f.write(' 21 21\n')
    f.write(' 1 0 0 2\n')
    f.write(' 0 1 0 2\n')
    f.write('END\n\n')

    f.write('Orbitals GenBas\n')
    for orb_index in range(nao):
        f.write('  Basis%%AO%d\n' % orb_index)
    f.write('End\n\n')


class TestAOvaluesADF(unittest.TestCase):

    def setUp(self):

        # define the molecule
        path_hdf5 = (
            PATH_TEST / 'hdf5/C_adf_dzp.hdf5').absolute().as_posix()
        self.mol = Molecule(load=path_hdf5)

        # define the wave function
        self.wf = SlaterJastrow(self.mol, include_all_mo=True)

        # define the grid points
        self.npts = 21
        pts = get_pts(self.npts)

        self.pos = torch.zeros(self.npts**2, self.mol.nelec * 3)
        self.pos[:, :3] = pts
        self.pos = Variable(self.pos)
        self.pos.requires_grad = True

    def test_ao(self):

        aovals = self.wf.ao(self.pos).detach().numpy()

        for iorb in range(self.mol.basis.nao):

            path_cube = PATH_TEST / f'cube/C_AO_%Basis%AO{iorb}.cub'
            fname = path_cube.absolute().as_posix()
            adf_ref_data = np.array(read_cubefile(
                fname)).reshape(self.npts, self.npts)
            qmctorch_data = (aovals[:, 0, iorb]).reshape(
                self.npts, self.npts)

            delta = np.abs(adf_ref_data - qmctorch_data)

            if __PLOT__:
                plt.subplot(1, 3, 1)
                plt.imshow(adf_ref_data)

                plt.subplot(1, 3, 2)
                plt.imshow(qmctorch_data)

                plt.subplot(1, 3, 3)
                plt.imshow(delta)
                plt.show()

            assert(delta.mean() < 1E-3)


if __name__ == "__main__":
    unittest.main()
    # t = TestAOvaluesADF()
    # t.setUp()
    # t.test_ao()
    # t.test_mo()



import torch
from ...utils import fast_power


class Harmonics:

    def __init__(self, type, **kwargs):
        """Compute spherical or cartesian harmonics and their derivatives

        Args:
            type (str): harmonics type (cart or sph)

        Keyword Arguments:
            bas_l (torch.tensor): second quantum numbers (sph)
            bas_m (torch.tensor): third quantum numbers (sph)
            bas_kx (torch.tensor): x exponent (cart)
            bas_ky (torch.tensor): xy exponent (cart)
            bas_kz (torch.tensor): z exponent (cart)
            cuda (bool): use cuda (defaults False)

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> harm = Harmonics(cart)
            >>> pos = torch.rand(100,6)
            >>> hvals = harm(pos)
            >>> dhvals = harm(pos,derivative=1)
        """

        self.type = type

        # check if we need cuda
        if 'cuda' not in kwargs:
            cuda = False
        else:
            cuda = kwargs['cuda']

        # select the device
        if cuda:
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

        # register parameters
        if self.type == 'sph':
            self.bas_l = torch.as_tensor(
                kwargs['bas_l']).to(self.device)
            self.bas_m = torch.as_tensor(
                kwargs['bas_m']).to(self.device)

        elif self.type == 'cart':

            self.bas_kx = torch.as_tensor(
                kwargs['bas_kx']).to(self.device)
            self.bas_ky = torch.as_tensor(
                kwargs['bas_ky']).to(self.device)
            self.bas_kz = torch.as_tensor(
                kwargs['bas_kz']).to(self.device)

            self.bas_k = torch.stack(
                (self.bas_kx, self.bas_ky, self.bas_kz)).transpose(0, 1)
            self.mask_bas_k0 = self.bas_k == 0
            self.mask_bas_k2 = self.bas_k == 2

    def __call__(self, xyz, derivative=[0], sum_grad=True, sum_hess=True):
        """Computes the cartesian or spherical harmonics

        Arguments:
            xyz {torch.tensor} -- coordinate of each electrons from each BAS
                                  center (Nbatch, Nelec, Nbas, Ndim)

        Keyword Arguments:
            derivative {int} -- order of the derivative (default: {0})
            sum_grad {bool} -- return the sum of th derivative if true and
                               grad if False (default: {True})
            sum_hess {bool} -- return the sum of the 2nd derivative if true and
                               grad if False (default: {True})

        Raises:
            ValueError: of type is unrecognized

        Returns:
            torch.tensor -- Values or gradient of the spherical harmonics
        """

        if self.type == 'cart':
            return CartesianHarmonics(xyz, self.bas_k, self.mask_bas_k0, self.mask_bas_k2,
                                      derivative, sum_grad, sum_hess)
        elif self.type == 'sph':
            return SphericalHarmonics(
                xyz, self.bas_l, self.bas_m, derivative, sum_grad, sum_hess)
        else:
            raise ValueError('Harmonics type should be cart or sph')


def CartesianHarmonics(xyz, k, mask0, mask2, derivative=[0],
                       sum_grad=True, sum_hess=True):
    r"""Computes Real Cartesian Harmonics

    .. math::
        Y = x^{k_x} \\times y^{k_y} \\times z^{k_z}

    Args:
        xyz (torch.tensor): distance between sampling points and orbital centers \n
                            size : (Nbatch, Nelec, Nbas, Ndim)
        k (torch.tensor): (kx,ky,kz) exponents
        mask0 (torch.tensor): precomputed mask of k=0
        mask2 (torch.tensor): precomputed mask of k=2
        derivative (int, optional): degree of the derivative. Defaults to 0.
        sum_grad (bool, optional): returns the sum of the derivative if True. Defaults to True.
        sum_hess (bool, optional): returns the sum of the 2nd derivative if True. Defaults to True.
    Returns:
        torch.tensor: values of the harmonics at the sampling points
    """

    if not isinstance(derivative, list):
        derivative = [derivative]

    def _kernel():
        return xyz_k.prod(-1)

    def _first_derivative_kernel():
        km1 = k-1
        km1[km1 < 0] = 0

        xyz_km1 = fast_power(xyz, km1)

        kx, ky, kz = k.transpose(0, 1)
        dx = kx * xyz_km1[..., 0] * xyz_k[..., 1] * xyz_k[..., 2]
        dy = ky * xyz_k[..., 0] * xyz_km1[..., 1] * xyz_k[..., 2]
        dz = kz * xyz_k[..., 0] * xyz_k[..., 1] * xyz_km1[..., 2]

        if sum_grad:
            return dx + dy + dz
        else:
            return torch.stack((dx, dy, dz), dim=-1)

    def _second_derivative_kernel():
        # prepare the exponets
        km2 = k - 2
        km2[km2 < 0] = 0

        xyz_km2 = fast_power(xyz, km2)

        kx, ky, kz = k.transpose(0, 1)

        d2x = kx*(kx-1) * xyz_km2[..., 0] * \
            xyz_k[..., 1] * xyz_k[..., 2]
        d2y = ky*(ky-1) * xyz_k[..., 0] * \
            xyz_km2[..., 1] * xyz_k[..., 2]
        d2z = kz*(kz-1) * xyz_k[..., 0] * \
            xyz_k[..., 1] * xyz_km2[..., 2]

        if sum_hess:
            return d2x + d2y + d2z
        else:
            return torch.stack((d2x, d2y, d2z), dim=-1)

    def _mixed_second_derivative_kernel():
        km1 = k-1
        km1[km1 < 0] = 0

        xyz_km1 = fast_power(xyz, km1)

        kx, ky, kz = k.transpose(0, 1)

        dxdy = kx * xyz_km1[..., 0] * ky * \
            xyz_km1[..., 1] * xyz_k[..., 2]
        dxdz = kx * xyz_km1[..., 0] * \
            xyz_k[..., 1] * kz * xyz_km1[..., 2]
        dydz = xyz_k[..., 0] * ky * \
            xyz_km1[..., 1] * kz * xyz_km1[..., 2]

        return torch.stack((dxdy, dxdz, dydz), dim=-1)

    # computes the power of the xyz
    xyz_k = fast_power(xyz, k,  mask0, mask2)

    # compute the outputs
    fns = [_kernel,
           _first_derivative_kernel,
           _second_derivative_kernel,
           _mixed_second_derivative_kernel]

    output = []
    for d in derivative:
        output.append(fns[d]())

    if len(derivative) == 1:
        return output[0]
    else:
        return output


def SphericalHarmonics(xyz, l, m, derivative=0, sum_grad=True, sum_hess=True):
    r"""Compute the Real Spherical Harmonics of the AO.

    Args:
        xyz (torch.tensor): distance between sampling points and orbital centers \n
                            size : (Nbatch, Nelec, Nbas, Ndim)
        l (torch.tensor):  l quantum number
        m (torch.tensor):  m quantum number

    Returns:
        Y (torch.tensor): value of each harmonics at each points (or derivative) \n
                          size : (Nbatch,Nelec,Nrbf) for sum_grad=True \n
                          size : (Nbatch,Nelec,Nrbf, Ndim) for sum_grad=False
    """
    if not sum_hess:
        raise NotImplementedError(
            'SphericalHarmonics cannot return individual component of the laplacian')
    if derivative > 2:
        raise NotImplementedError(
            "Spherical Harmonics only accpet derivative=0,1,2 (%d found)" % derivative)

    if sum_grad:
        return get_spherical_harmonics(xyz, l, m, derivative)
    else:
        if derivative != 1:
            raise ValueError(
                'Gradient of the spherical harmonics require derivative=1')
        return get_grad_spherical_harmonics(xyz, l, m)


def get_spherical_harmonics(xyz, lval, m, derivative):
    r"""Compute the Real Spherical Harmonics of the AO.

    Args:
        xyz (torch.tensor): distance between sampling points and orbital centers \n
                            size : (Nbatch, Nelec, Nbas, Ndim)
        l (torch.tensor): l quantum number
        m (torch.tensor): m quantum number

    Returns:
        Y (torch.tensor): value of each harmonics at each points (or derivative) \n
                          size : (Nbatch,Nelec,Nrbf)
    """

    Y = torch.zeros_like(xyz[..., 0])

    # l=0
    ind = (lval == 0).nonzero().view(-1)
    if derivative == 0:
        Y[:, :, ind] = _spherical_harmonics_l0(xyz[:, :, ind, :])
    if derivative == 1:
        Y[:, :, ind] = _nabla_spherical_harmonics_l0(
            xyz[:, :, ind, :])

    # l=1
    indl = (lval == 1)
    if torch.any(indl):
        for mval in [-1, 0, 1]:
            indm = (m == mval)
            ind = (indl * indm).nonzero().view(-1)
            if len(ind > 0):
                if derivative == 0:
                    Y[:, :, ind] = _spherical_harmonics_l1(
                        xyz[:, :, ind, :], mval)
                if derivative == 1:
                    Y[:, :, ind] = _nabla_spherical_harmonics_l1(
                        xyz[:, :, ind, :], mval)
                if derivative == 2:
                    Y[:, :, ind] = _lap_spherical_harmonics_l1(
                        xyz[:, :, ind, :], mval)

    # l=2
    indl = (lval == 2)
    if torch.any(indl):
        for mval in [-2, -1, 0, 1, 2]:
            indm = (m == mval)
            ind = (indl * indm).nonzero().view(-1)
            if len(ind > 0):
                if derivative == 0:
                    Y[:, :, ind] = _spherical_harmonics_l2(
                        xyz[:, :, ind, :], mval)
                if derivative == 1:
                    Y[:, :, ind] = _nabla_spherical_harmonics_l2(
                        xyz[:, :, ind, :], mval)
                if derivative == 2:
                    Y[:, :, ind] = _lap_spherical_harmonics_l2(
                        xyz[:, :, ind, :], mval)

    return Y


def get_grad_spherical_harmonics(xyz, lval, m):
    r"""Compute the gradient of the Real Spherical Harmonics of the AO.

    Args:
        xyz (torch.tensor): distance between sampling points and orbital centers \n
                            size : (Nbatch, Nelec, Nbas, Ndim)
        l (torch.tensor): l quantum number
        m (torch.tensor): m quantum number

    Returns:
        Y (torch.tensor): value of each harmonics at each points (or derivative) \n
                          size : (Nbatch,Nelec,Nrbf,3)
    """

    Y = torch.zeros_like(xyz)

    # l=0
    ind = (lval == 0).nonzero().view(-1)
    Y[:, :, ind, :] = _grad_spherical_harmonics_l0(xyz[:, :, ind, :])

    # l=1
    indl = (lval == 1)
    if torch.any(indl):
        for mval in [-1, 0, 1]:
            indm = (m == mval)
            ind = (indl * indm).nonzero().view(-1)
            if len(ind > 0):
                # _tmp = _grad_spherical_harmonics_l1(xyz[:, :, ind, :], mval)
                Y[:, :, ind, :] = _grad_spherical_harmonics_l1(
                    xyz[:, :, ind, :], mval)

    # l=2
    indl = (lval == 2)
    if torch.any(indl):
        for mval in [-2, -1, 0, 1, 2]:
            indm = (m == mval)
            ind = (indl * indm).nonzero().view(-1)
            if len(ind > 0):
                Y[:, :, ind, :] = _grad_spherical_harmonics_l2(
                    xyz[:, :, ind, :], mval)

    return Y

# =============== L0


def _spherical_harmonics_l0(xyz):
    r"""Compute the l=0 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
    Returns
        Y00 = 1/2 \sqrt(1 / \pi)
    """

    return 0.2820948 * torch.ones_like(xyz[..., 0])


def _nabla_spherical_harmonics_l0(xyz):
    r"""Compute the nabla of l=0 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
    Returns
        \nabla * Y00 = 0
    """
    return torch.zeros_like(xyz[..., 0])


def _grad_spherical_harmonics_l0(xyz):
    r"""Compute the nabla of l=0 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
    Returns
        \nabla * Y00 = 0
    """
    return torch.zeros_like(xyz)


def _lap_spherical_harmonics_l0(xyz):
    r"""Compute the laplacian of l=0 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
    Returns
        \nabla^2 * Y00 = 0
    """
    return torch.zeros_like(xyz[..., 0])

# =============== L1


def _spherical_harmonics_l1(xyz, m):
    r"""Compute the 1-1 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-1,0,1)
    Returns
        Y0-1 = \sqrt(3 / (4\pi)) y/r. (m=-1)
        Y00  = \sqrt(3 / (4\pi)) z/r (m=0)
        Y01  = \sqrt(3 / (4\pi)) x/r. (m=1)
    """
    index = {-1: 1, 0: 2, 1: 0}
    r = torch.sqrt((xyz**2).sum(3))
    c = 0.4886025119029199
    return c * xyz[:, :, :, index[m]] / r


def _nabla_spherical_harmonics_l1(xyz, m):
    r"""Compute the nabla of 1-1 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-1,0,1)
    Returns
        \nabla Y0-1 = \sqrt(3 / (4\pi)) ( 1/r - y (x+y+z)/r^3 ) (m=-1)
        \nabla Y00  = \sqrt(3 / (4\pi)) ( 1/r - z (x+y+z)/r^3 ) (m= 0)
        \nabla Y01  = \sqrt(3 / (4\pi)) ( 1/r - x (x+y+z)/r^3 ) (m= 1)
    """
    index = {-1: 1, 0: 2, 1: 0}
    r = torch.sqrt((xyz**2).sum(3))
    r3 = r**3
    c = 0.4886025119029199
    return c * (1. / r - xyz[:, :, :, index[m]] * xyz.sum(3) / r3)


def _grad_spherical_harmonics_l1(xyz, m):
    r"""Compute the nabla of 1-1 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-1,0,1)
    Returns
        \nabla Y0-1 = \sqrt(3 / (4\pi)) ( 1/r^3 * [-yx, x^2+z^2, -yz]  ) (m=-1)
        \nabla Y00  = \sqrt(3 / (4\pi)) ( 1/r^3 * [-zx, -zy, x^2+y^2]  ) (m= 0)
        \nabla Y01  = \sqrt(3 / (4\pi)) ( 1/r^3 * [y^2+z^2, -xy, -xz]  ) (m=-1)
    """

    r = torch.sqrt((xyz**2).sum(3))
    r3 = r**3
    c = 0.4886025119029199
    p = (c / r3).unsqueeze(-1)

    if m == -1:
        return p * (torch.stack([-xyz[:, :, :, 1] * xyz[:, :, :, 0],
                                 xyz[:, :, :, 0]**2 +
                                 xyz[:, :, :, 2]**2,
                                 -xyz[:, :, :, 1] * xyz[:, :, :, 2]],
                                dim=-1))
    if m == 0:

        return p * (torch.stack([-xyz[:, :, :, 2] * xyz[:, :, :, 0],
                                 -xyz[:, :, :, 2] * xyz[:, :, :, 1],
                                 xyz[:, :, :, 0]**2 + xyz[:, :, :, 1]**2],
                                dim=-1))
    if m == 1:
        return p * (torch.stack([xyz[:, :, :, 1]**2 + xyz[:, :, :, 2]**2,
                                 -xyz[:, :, :, 0] * xyz[:, :, :, 1],
                                 -xyz[:, :, :, 0] * xyz[:, :, :, 2]],
                                dim=-1))


def _lap_spherical_harmonics_l1(xyz, m):
    r"""Compute the laplacian of 1-1 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-1,0,1)
    Returns
        Y0-1 = \sqrt(3 / (4\pi)) ( -2y/r^3 ) (m=-1)
        Y00  = \sqrt(3 / (4\pi)) ( -2z/r^3 ) (m= 0)
        Y01  = \sqrt(3 / (4\pi)) ( -2x/r^3 ) (m= 1)
    """
    index = {-1: 1, 0: 2, 1: 0}
    r = torch.sqrt((xyz**2).sum(3))
    r3 = r**3
    c = 0.4886025119029199
    return c * (- 2 * xyz[:, :, :, index[m]] / r3)

# =============== L2


def _spherical_harmonics_l2(xyz, m):
    r"""Compute the l=2 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-2,-1,0,1,2)
    Returns
        Y2-2 = 1/2\sqrt(15/\pi) xy/r^2
        Y2-1 = 1/2\sqrt(15/\pi) yz/r^2
        Y20  = 1/4\sqrt(5/\pi) (-x^2-y^2+2z^2)/r^2
        Y21  = 1/2\sqrt(15/\pi) zx/r^2
        Y22  = 1/4\sqrt(15/\pi) (x*x-y*y)/r^2
    """

    r2 = (xyz**2).sum(-1)

    if m == 0:
        c0 = 0.31539156525252005
        return c0 * (-xyz[:, :, :, 0]**2 - xyz[:, :, :, 1]
                     ** 2 + 2 * xyz[:, :, :, 2]**2) / r2
    if m == 2:
        c2 = 0.5462742152960396
        return c2 * (xyz[:, :, :, 0]**2 - xyz[:, :, :, 1]**2) / r2
    else:
        cm = 1.0925484305920792
        index = {-2: [0, 1], -1: [1, 2], 1: [2, 0]}
        return cm * xyz[:, :, :, index[m][0]] * \
            xyz[:, :, :, index[m][1]] / r2


def _nabla_spherical_harmonics_l2(xyz, m):
    r"""Compute the nabla of l=2 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-2,-1,0,1,2)
    Returns
        Y2-2 = 1/2\sqrt(15/\pi) (x+y)/r^2 - 2 * xy (x+y+z)/r^3
        Y2-1 = 1/2\sqrt(15/\pi) (y+z)/r^2 - 2 * yz (x+y+z)/r^3
        Y20  = 1/4\sqrt(5/\pi) ( (-2x-2y+4z)/r^2 - \
               2 *(-xx - yy + 2zz) * (x+y+z)/r3 )
        Y21  = 1/2\sqrt(15/\pi) (x+z)/r^2 - 2 * xz (x+y+z)/r^3
        Y22  = 1/4\sqrt(15/\pi)  ( 2(x-y)/r^2 - 2 *(xx-yy)(x+y+z)/r^3  )
    """

    r = torch.sqrt((xyz**2).sum(3))
    r2 = r**2
    r3 = r**3

    if m == 0:
        c0 = 0.31539156525252005
        return c0 * ((- 2 * xyz[:, :, :, 0] - 2 * xyz[:, :, :, 1] + 4 * xyz[:, :, :, 2]) / r2
                     - 2 * (-xyz[:, :, :, 0]**2 - xyz[:, :, :, 1]**2 + 2 * xyz[:, :, :, 2]**2) * xyz.sum(3) / r3)
    if m == 2:
        c2 = 0.5462742152960396
        return c2 * (2 * (xyz[:, :, :, 0] - xyz[:, :, :, 1]) / r2 - 2 * (xyz[:, :, :, 0]**2
                                                                         - xyz[:, :, :, 1]**2) * xyz.sum(3) / r3)
    else:
        cm = 1.0925484305920792
        index = {-2: [0, 1], -1: [1, 2], 1: [2, 0]}
        return cm * ((xyz[:, :, :, index[m][0]] + xyz[:, :, :, index[m][1]]) / r2
                     - 2 * xyz[:, :, :, index[m][0]] * xyz[:, :, :, index[m][1]] * xyz.sum(3) / r3)


def _grad_spherical_harmonics_l2(xyz, m):
    r"""Compute the nabla of l=2 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-2,-1,0,1,2)

    Returns
        Y2-2 = 1/2 \sqrt(15/\pi) 1./r4 ([y(-xx+yy+zz, x(-yy+xx+zz,-2xyz))])
        Y2-1 = 1/2 \sqrt(15/\pi)
        Y20  = 1/4 \sqrt(5/\pi)
        Y21  = 1/2 \sqrt(15/\pi)
        Y22  = 1/4 \sqrt(15/\pi)
    """

    r = torch.sqrt((xyz**2).sum(3))
    r4 = r**4

    x = xyz[:, :, :, 0]
    y = xyz[:, :, :, 1]
    z = xyz[:, :, :, 2]

    if m == -2:
        c0 = 0.31539156525252005
        p = (c0 / r4).unsqueeze(-1)
        return p * (torch.stack([y * (-x**2 + y**2 + z**2),
                                 x * (-y**2 + x**2 + z**2),
                                 -2 * xyz.prod(-1)],
                                dim=-1))
    if m == -1:
        c0 = 0.31539156525252005
        p = (c0 / r4).unsqueeze(-1)
        return p * (torch.stack([-2 * xyz.prod(-1),
                                 z * (-y**2 + x**2 + z**2),
                                 y * (-z**2 + x**2 + y**2)],
                                dim=-1))
    if m == 0:
        c0 = 0.31539156525252005
        p = (c0 / r4).unsqueeze(-1)
        return p * (torch.stack([-6 * x * z * z,
                                 -6 * y * z * z,
                                 6 * x * x * z + 6 * y * y * z],
                                dim=-1))

    if m == 1:
        c0 = 0.31539156525252005
        p = (c0 / r4).unsqueeze(-1)
        return p * (torch.stack([z * (-x * x + y * y + z * z),
                                 -2 * xyz.prod(-1),
                                 x * (x * x + y * y - z * z)],
                                dim=-1))
    if m == 2:
        c0 = 0.5462742152960396
        p = (c0 / r4).unsqueeze(-1)
        return p * (torch.stack([4 * x * y * y + 2 * x * z * z,
                                 -4 * x * x * y - 2 * y * z * z,
                                 -2 * z * (x * x - y * y)],
                                dim=-1))


def _lap_spherical_harmonics_l2(xyz, m):
    r"""Compute the nabla of l=2 Spherical Harmonics
    Args:
        xyz : array (Nbatch,Nelec,Nrbf,Ndim) x,y,z, of (Point - Center)
        m : second quantum number (-2,-1,0,1,2)
    Returns
        Y2-2 = 1/2\sqrt(15/\pi) -6xy/r^4
        Y2-1 = 1/2\sqrt(15/\pi) -6yz/r^4

        Y20  = 1/4\sqrt(5/\pi) ( 6/r6 * (xx+yy)^2 - zz * (xx + yy -2zz))

        Y21  = 1/2\sqrt(15/\pi) -6zx/r^4
        Y22  = 1/4\sqrt(15/\pi)  ( 6/r6 * ( zz*(yy-xx)  +y^4 - x^4  )  )
    """

    r = torch.sqrt((xyz**2).sum(3))
    r4 = r**4
    r6 = r**6

    if m == 0:
        c0 = 0.31539156525252005
        xyz2 = xyz**2
        return c0 * (6 / r6 * (xyz2[:, :, :, :2].sum(-1))**2 - xyz2[:, :, :, 2] * (xyz2[:, :, :, 0]
                                                                                   + xyz2[:, :, :, 1] - 2 * xyz2[:, :, :, 2]))
    if m == 2:
        c2 = 0.5462742152960396
        xyz2 = xyz**2
        return c2 * (6 / r6 * xyz2[:, :, :, 2] * (xyz2[:, :, :, 1] - xyz2[:, :, :, 0])
                     + xyz2[:, :, :, 1]**2 - xyz2[:, :, :, 0]**2)
    else:
        cm = 1.0925484305920792
        index = {-2: [0, 1], -1: [1, 2], 1: [2, 0]}
        return cm * (- 6 * xyz[:, :, :, index[m][0]]
                     * xyz[:, :, :, index[m][1]] / r4)




import torch
from torch import nn
from functools import reduce

from .elec_elec.jastrow_factor_electron_electron import JastrowFactorElectronElectron
from .elec_nuclei.jastrow_factor_electron_nuclei import JastrowFactorElectronNuclei
from .elec_elec_nuclei.jastrow_factor_electron_electron_nuclei import JastrowFactorElectronElectronNuclei


from .elec_elec.kernels.pade_jastrow_kernel import PadeJastrowKernel as PadeJastrowKernelElecElec
from .elec_nuclei.kernels.pade_jastrow_kernel import PadeJastrowKernel as PadeJastrowKernelElecNuc


class JastrowFactorCombinedTerms(nn.Module):

    def __init__(self, nup, ndown, atomic_pos,
                 jastrow_kernel={
                     'ee': PadeJastrowKernelElecElec,
                     'en': PadeJastrowKernelElecNuc,
                     'een': None},
                 jastrow_kernel_kwargs={
                     'ee': {},
                     'en': {},
                     'een': {}},
                 cuda=False):
        """[summary]

        Args:
            nup (int): number of spin up electron
            ndown (int): number opf spin down electron
            atomic_pos (torch tensor): atomic positions
            jastrow_kernel ([dict]): kernels of the jastrow factor
            jastrow_kernel_kwargs (dict): keyword argument of the kernels
            cuda (bool, optional): [description]. Defaults to False.
        """

        super().__init__()
        self.nup = nup
        self.ndown = ndown
        self.cuda = cuda

        self.jastrow_terms = []

        # sanitize the dict
        for k in ['ee', 'en', 'een']:
            if k not in jastrow_kernel.keys():
                jastrow_kernel[k] = None
            if k not in jastrow_kernel_kwargs.keys():
                jastrow_kernel_kwargs[k] = {}

        self.requires_autograd = True

        if jastrow_kernel['ee'] is not None:

            self.jastrow_terms.append(JastrowFactorElectronElectron(nup, ndown,
                                                                    jastrow_kernel['ee'],
                                                                    jastrow_kernel_kwargs['ee'],
                                                                    cuda=cuda))

        if jastrow_kernel['en'] is not None:

            self.jastrow_terms.append(JastrowFactorElectronNuclei(nup, ndown,
                                                                  atomic_pos,
                                                                  jastrow_kernel['en'],
                                                                  jastrow_kernel_kwargs['en'],
                                                                  cuda=cuda))

        if jastrow_kernel['een'] is not None:

            self.jastrow_terms.append(JastrowFactorElectronElectronNuclei(nup, ndown,
                                                                          atomic_pos,
                                                                          jastrow_kernel['een'],
                                                                          jastrow_kernel_kwargs['een'],
                                                                          cuda=cuda))
        self.nterms = len(self.jastrow_terms)

    def forward(self, pos, derivative=0, sum_grad=True):
        """Compute the Jastrow factors.

        Args:
            pos(torch.tensor): Positions of the electrons
               Size: Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0, 1, 2,).
              Defaults to 0.
            sum_grad(bool, optional): Return the sum_grad(i.e. the sum of
                                                           the derivatives)
              terms. Defaults to True.
                False only for derivative = 1

        Returns:
            torch.tensor: value of the jastrow parameter for all confs
              derivative = 0  (Nmo) x Nbatch x 1
               derivative = 1  (Nmo) x Nbatch x Nelec
                  (for sum_grad = True)
                derivative = 1  (Nmo) x Nbatch x Ndim x Nelec
                  (for sum_grad = False)
        """
        if derivative == 0:

            jast_vals = [term(pos) for term in self.jastrow_terms]
            return self.get_combined_values(jast_vals)

        elif derivative == 1:

            if sum_grad:
                jast_vals = [term(pos) for term in self.jastrow_terms]
            else:
                jast_vals = [term(pos).unsqueeze(-1)
                             for term in self.jastrow_terms]
            djast_vals = [term(pos, derivative=1, sum_grad=sum_grad)
                          for term in self.jastrow_terms]

            return self.get_derivative_combined_values(jast_vals, djast_vals)

        elif derivative == 2:

            jast_vals = [term(pos)
                         for term in self.jastrow_terms]
            djast_vals = [term(pos, derivative=1, sum_grad=False)
                          for term in self.jastrow_terms]
            d2jast_vals = [term(pos, derivative=2)
                           for term in self.jastrow_terms]
            return self.get_second_derivative_combined_values(jast_vals, djast_vals, d2jast_vals)

        elif derivative == [0, 1, 2]:

            jast_vals = [term(pos) for term in self.jastrow_terms]
            djast_vals = [term(pos, derivative=1, sum_grad=False)
                          for term in self.jastrow_terms]
            d2jast_vals = [term(pos, derivative=2)
                           for term in self.jastrow_terms]

            # combine the jastrow terms
            out_jast = self.get_combined_values(jast_vals)

            # combine the second derivative
            out_d2jast = self.get_second_derivative_combined_values(
                jast_vals, djast_vals, d2jast_vals)

            # unsqueeze the jast terms to be compatible with the
            # derivative
            jast_vals = [j.unsqueeze(-1) for j in jast_vals]

            # combine the derivative
            out_djast = self.get_derivative_combined_values(
                jast_vals, djast_vals)

            return(out_jast, out_djast, out_d2jast)

        else:
            raise ValueError('derivative not understood')

    @ staticmethod
    def get_combined_values(jast_vals):
        """Compute the product of all terms in jast_vals."""
        if len(jast_vals) == 1:
            return jast_vals[0]
        else:
            return reduce(lambda x, y: x*y, jast_vals)

    @ staticmethod
    def get_derivative_combined_values(jast_vals, djast_vals):
        """Compute the derivative of the product.
        .. math:
            J = A * B * C
            \\frac{d J}{dx} = \\frac{d A}{dx} B C + A \\frac{d B}{dx} C + A B \\frac{d C}{dx}
        """
        if len(djast_vals) == 1:
            return djast_vals[0]
        else:
            out = 0.
            nterms = len(jast_vals)
            for i in range(nterms):
                tmp = jast_vals.copy()
                tmp[i] = djast_vals[i]
                out += reduce(lambda x, y: x*y, tmp)
            return out

    @ staticmethod
    def get_second_derivative_combined_values(jast_vals, djast_vals, d2jast_vals):
        """Compute the derivative of the product.
        .. math:
            J = A * B * C
            \\frac{d^2 J}{dx^2} = \\frac{d^2 A}{dx^2} B C + A \\frac{d^2 B}{dx^2} C + A B \\frac{d^2 C}{dx^2} \\
                               + 2( \\frac{d A}{dx} \\frac{dB}{dx} C + \\frac{d A}{dx} B \\frac{dC}{dx} + A \\frac{d B}{dx} \\frac{dC}{dx} )
        """
        if len(d2jast_vals) == 1:
            return d2jast_vals[0]
        else:
            out = 0.
            nterms = len(jast_vals)
            for i in range(nterms):

                # d2a * b * c
                tmp = jast_vals.copy()
                tmp[i] = d2jast_vals[i]
                out = out + reduce(lambda x, y: x*y, tmp)

            for i in range(nterms-1):
                for j in range(i+1, nterms):

                    # da * db * c
                    tmp = jast_vals.copy()
                    tmp = [j.unsqueeze(-1) for j in tmp]
                    tmp[i] = djast_vals[i]
                    tmp[j] = djast_vals[j]

                    out = out + \
                        (2.*reduce(lambda x, y: x*y, tmp)).sum(1)

            return out



"""Module containing some utilities for testing."""

from pathlib import Path
import pkg_resources as pkg

__all__ = ["PATH_QMCTORCH", "PATH_TEST"]

# Environment data
PATH_QMCTORCH = Path(pkg.resource_filename('qmctorch', ''))
ROOT = PATH_QMCTORCH.parent

PATH_TEST = ROOT / "tests"





import numpy as np
import torch
from .slater_jastrow_base import SlaterJastrowBase

from .jastrows.elec_elec.kernels.pade_jastrow_kernel import PadeJastrowKernel
from .jastrows.elec_elec.jastrow_factor_electron_electron import JastrowFactorElectronElectron


class SlaterJastrow(SlaterJastrowBase):

    def __init__(self, mol, configs='ground_state',
                 kinetic='jacobi',
                 jastrow_kernel=PadeJastrowKernel,
                 jastrow_kernel_kwargs={},
                 cuda=False,
                 include_all_mo=True):
        """Slater Jastrow wave function with electron-electron Jastrow factor

        .. math::
            \\Psi(R_{at}, r) = J(r)\\sum_n c_n D^\\uparrow_n(r^\\uparrow)D^\\downarrow_n(r^\\downarrow)

        with

        .. math::
            J(r) = \\exp\\left( K_{ee}(r) \\right)

        with K, a kernel function depending only on the electron-eletron distances 

        Args:
            mol (Molecule): a QMCTorch molecule object
            configs (str, optional): defines the CI configurations to be used. Defaults to 'ground_state'.
                - ground_state : only the ground state determinant in the wave function
                - single(n,m) : only single excitation with n electrons and m orbitals 
                - single_double(n,m) : single and double excitation with n electrons and m orbitals
                - cas(n, m) : all possible configuration using n eletrons and m orbitals                   
            kinetic (str, optional): method to compute the kinetic energy. Defaults to 'jacobi'.
                - jacobi : use the Jacobi formula to compute the kinetic energy 
                - auto : use automatic differentiation to compute the kinetic energy
            jastrow_kernel (JastrowKernelBase, optional) : Class that computes the jastrow kernels
            jastrow_kernel_kwargs (dict, optional) : keyword arguments for the jastrow kernel contructor
            cuda (bool, optional): turns GPU ON/OFF  Defaults to Fals   e.
            include_all_mo (bool, optional): include either all molecular orbitals or only the ones that are
                                             popualted in the configs. Defaults to False
        Examples::
            >>> from qmctorch.scf import Molecule
            >>> from qmctorch.wavefunction import SlaterJastrow
            >>> mol = Molecule('h2o.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterJastrow(mol, configs='cas(2,2)')
        """

        super().__init__(mol, configs, kinetic, cuda, include_all_mo)

        # process the Jastrow
        if jastrow_kernel is not None:

            self.use_jastrow = True
            self.jastrow_type = jastrow_kernel.__name__
            self.jastrow = JastrowFactorElectronElectron(
                self.mol.nup, self.mol.ndown, jastrow_kernel,
                kernel_kwargs=jastrow_kernel_kwargs, cuda=cuda)

            if self.cuda:
                self.jastrow = self.jastrow.to(self.device)

        self.log_data()

    def forward(self, x, ao=None):
        """computes the value of the wave function for the sampling points

        .. math::
            \\Psi(R) =  J(R) \\sum_{n} c_n  D^{u}_n(r^u) \\times D^{d}_n(r^d)

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            ao (torch.tensor, optional): values of the atomic orbitals (Nbatch, Nelec, Nao)

        Returns:
            torch.tensor: values of the wave functions at each sampling point (Nbatch, 1)

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterJastrow(mol, configs='cas(2,2)')
            >>> pos = torch.rand(500,6)
            >>> vals = wf(pos)
        """

        if self.use_jastrow:
            J = self.jastrow(x)

        # atomic orbital
        if ao is None:
            x = self.ao(x)
        else:
            x = ao

        # molecular orbitals
        x = self.mo_scf(x)

        # mix the mos
        x = self.mo(x)

        # pool the mos
        x = self.pool(x)

        if self.use_jastrow:
            return J * self.fc(x)

        else:
            return self.fc(x)

    def ao2mo(self, ao):
        return self.mo(self.mo_scf(ao))

    def pos2mo(self, x, derivative=0):
        """Get the values of MOs

        Arguments:
            x {torch.tensor} -- positions of the electrons [nbatch, nelec*ndim]

        Keyword Arguments:
            derivative {int} -- order of the derivative (default: {0})

        Returns:
            torch.tensor -- MO matrix [nbatch, nelec, nmo]
        """
        return self.mo(self.mo_scf(self.ao(x, derivative=derivative)))

    def kinetic_energy_jacobi(self, x, **kwargs):
        """Compute the value of the kinetic enery using the Jacobi Formula.
        C. Filippi, Simple Formalism for Efficient Derivatives .

        .. math::
            \\frac{\Delta \\Psi(R)}{\\Psi(R)} = \\Psi(R)^{-1} \\sum_n c_n (\\frac{\Delta D_n^u}{D_n^u} + \\frac{\Delta D_n^d}{D_n^d}) D_n^u D_n^d

        We compute the laplacian of the determinants through the Jacobi formula

        .. math::
            \\frac{\\Delta \\det(A)}{\\det(A)} = Tr(A^{-1} \\Delta A)

        Here :math:`A = J(R) \\phi` and therefore :

        .. math::
            \\Delta A = (\\Delta J) D + 2 \\nabla J \\nabla D + (\\Delta D) J

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the kinetic energy at each sampling points
        """

        ao, dao, d2ao = self.ao(x, derivative=[0, 1, 2])
        mo = self.ao2mo(ao)
        bkin = self.get_kinetic_operator(x, ao, dao, d2ao, mo)

        kin = self.pool.operator(mo, bkin)
        psi = self.pool(mo)
        out = self.fc(kin * psi) / self.fc(psi)
        return out

    def gradients_jacobi(self, x, sum_grad=False, pdf=False):
        """Compute the gradients of the wave function (or density) using the Jacobi Formula
        C. Filippi, Simple Formalism for Efficient Derivatives.

        .. math::
             \\frac{K(R)}{\Psi(R)} = Tr(A^{-1} B_{grad})

        The gradients of the wave function

        .. math::
            \\Psi(R) = J(R) \\sum_n c_n D^{u}_n D^{d}_n = J(R) \\Sigma

        are computed following

        .. math::
            \\nabla \\Psi(R) = \\left( \\nabla J(R) \\right) \\Sigma + J(R) \\left(\\nabla \Sigma \\right)

        with

        .. math::

            \\nabla \\Sigma =  \\sum_n c_n (\\frac{\\nabla D^u_n}{D^u_n} + \\frac{\\nabla D^d_n}{D^d_n}) D^u_n D^d_n

        that we compute with the Jacobi formula as:

        .. math::

            \\nabla \\Sigma =  \\sum_n c_n (Tr( (D^u_n)^{-1} \\nabla D^u_n) + Tr( (D^d_n)^{-1} \\nabla D^d_n)) D^u_n D^d_n

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            pdf (bool, optional) : if true compute the grads of the density

        Returns:
            torch.tensor: values of the gradients wrt the walker pos at each sampling points
        """

        # compute the mo values
        mo = self.ao2mo(self.ao(x))

        # compute the gradient operator matrix
        grad_ao = self.ao(x, derivative=1, sum_grad=False)

        # compute the derivatives of the MOs
        dmo = self.ao2mo(grad_ao.transpose(2, 3)).transpose(2, 3)
        dmo = dmo.permute(3, 0, 1, 2)

        # stride the tensor
        eye = torch.eye(self.nelec).to(self.device)
        dmo = dmo.unsqueeze(2) * eye.unsqueeze(-1)

        # reorder to have Nelec, Ndim, Nbatch, Nelec, Nmo
        dmo = dmo.permute(2, 0, 1, 3, 4)

        # flatten to have Nelec*Ndim, Nbatch, Nelec, Nmo
        dmo = dmo.reshape(-1, *(dmo.shape[2:]))

        # use the Jacobi formula to compute the value
        # the grad of each determinants and sum up the terms :
        # Tr( (D^u_n)^-1 \\nabla D^u_n) + Tr( (D^d_n)^-1 \\nabla D^d_n)
        grad_dets = self.pool.operator(mo, dmo)

        # compute the determinants
        # D^u_n D^d_n
        dets = self.pool(mo)

        # assemble the final values of \nabla \Sigma
        # \\sum_n c_n (Tr( (D^u_n)^-1 \\nabla D^u_n) + Tr( (D^d_n)^-1 \\nabla D^d_n)) D^u_n D^d_n
        out = self.fc(grad_dets * dets)
        out = out.transpose(0, 1).squeeze()

        if self.use_jastrow:

            nbatch = x.shape[0]

            # nbatch x 1
            jast = self.jastrow(x)

            # nbatch x ndim x nelec
            grad_jast = self.jastrow(x, derivative=1, sum_grad=False)

            # reorder grad_jast to nbtach x Nelec x Ndim
            grad_jast = grad_jast.permute(0, 2, 1)

            # compute J(R) (\nabla\Sigma)
            out = jast*out

            # add the product (\nabla J(R)) \Sigma
            out = out + \
                (grad_jast * self.fc(dets).unsqueeze(-1)).reshape(nbatch, -1)

        # compute the gradient of the pdf (i.e. the square of the wave function)
        # \nabla f^2 = 2 (\nabla f) f
        if pdf:
            out = 2 * out * self.fc(dets)
            if self.use_jastrow:
                out = out * jast

        return out

    def get_kinetic_operator(self, x, ao, dao, d2ao,  mo):
        """Compute the Bkin matrix

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            mo (torch.tensor, optional): precomputed values of the MOs

        Returns:
            torch.tensor: matrix of the kinetic operator
        """

        bkin = self.ao2mo(d2ao)

        if self.use_jastrow:

            jast, djast, d2jast = self.jastrow(x,
                                               derivative=[0, 1, 2],
                                               sum_grad=False)

            djast = djast.transpose(1, 2) / jast.unsqueeze(-1)
            d2jast = d2jast / jast

            dmo = self.ao2mo(dao.transpose(2, 3)).transpose(2, 3)

            djast_dmo = (djast.unsqueeze(2) * dmo).sum(-1)
            d2jast_mo = d2jast.unsqueeze(-1) * mo

            bkin = bkin + 2 * djast_dmo + d2jast_mo

        return -0.5 * bkin



import torch
from torch import nn

from .....utils import register_extra_attributes
from .jastrow_kernel_electron_electron_base import JastrowKernelElectronElectronBase


class PadeJastrowKernel(JastrowKernelElectronElectronBase):

    def __init__(self, nup, ndown, cuda, w=1.):
        """Computes the Simple Pade-Jastrow factor

        .. math::
            B_{ij} = \\frac{w_0 r_{ij}}{1 + w r_{ij}}

        where :math:`w_0` equals 0.5 for parallel spin and 0.25 for antiparallel spin

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            cuda (bool): Turns GPU ON/OFF.
            w (float, optional): Value of the variational parameter.
                                 Defaults to 1.

        """

        super().__init__(nup, ndown, cuda)

        self.weight = nn.Parameter(torch.as_tensor([w]),
                                   requires_grad=True)
        register_extra_attributes(self, ['weight'])

        self.static_weight = self.get_static_weight()
        self.requires_autograd = False

    def get_static_weight(self):
        """Get the matrix of static weights

        Returns:
            torch.tensor: matrix of the static weights
        """

        bup = torch.cat((0.25 * torch.ones(self.nup, self.nup), 0.5 *
                         torch.ones(self.nup, self.ndown)), dim=1)

        bdown = torch.cat((0.5 * torch.ones(self.ndown, self.nup), 0.25 *
                           torch.ones(self.ndown, self.ndown)), dim=1)

        static_weight = torch.cat((bup, bdown), dim=0).to(self.device)

        mask_tri_up = torch.triu(torch.ones_like(
            static_weight), diagonal=1).type(torch.BoolTensor).to(self.device)
        static_weight = static_weight.masked_select(mask_tri_up)

        return static_weight

    def forward(self, r):
        """ Get the jastrow kernel.

        .. math::
            B_{ij} = \\frac{w_0 r_{i,j}}{1+w r_{i,j}}

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: matrix of the jastrow kernels
                          Nbatch x Nelec x Nelec
        """
        return self.static_weight * r / (1.0 + self.weight * r)

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons

        .. math::

            \\frac{d B_{ij}}{d k_i} =  \\frac{d B_{ij}}{ d k_j }  = - \\frac{d B_{ji}}{d k_i}

        .. math::
            \\text{out}_{k,i,j} = A1 + A2

        .. math::
            A1_{kij} = w0 \\frac{dr_{ij}}{dk_i}  \\frac{1}{1 + w r_{ij}}

        .. math::
            A2_{kij} = - w0 w' r_{ij} \\frac{dr_{ij}}{dk_i} \\frac{1}{1 + w r_{ij}}^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        r_ = r.unsqueeze(1)
        denom = 1. / (1.0 + self.weight * r_)
        a = self.static_weight * dr * denom
        b = -self.static_weight * self.weight * r_ * dr * denom**2

        return (a + b)

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron

        .. math ::

            \\frac{d^2 B_{ij}}{d k_i^2} =  \\frac{d^2 B_{ij}}{d k_j^2} = \\frac{d^2 B_{ji}}{ d k_i^2}

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
            d2r (torch.tensor): matrix of the 2nd derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of
                          the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        r_ = r.unsqueeze(1)
        denom = 1. / (1.0 + self.weight * r_)
        denom2 = denom**2
        dr_square = dr * dr

        a = self.static_weight * d2r * denom
        b = -2 * self.static_weight * self.weight * dr_square * denom2
        c = -self.static_weight * self.weight * r_ * d2r * denom2
        d = 2 * self.static_weight * self.weight**2 * r_ * dr_square * denom**3

        return a + b + c + d



import torch
from torch import nn
from .backflow_kernel_base import BackFlowKernelBase


class BackFlowKernelSquare(BackFlowKernelBase):

    def __init__(self, mol, cuda=False):
        """Define a generic kernel to test the auto diff features."""
        super().__init__(mol, cuda)
        eps = 1E-4
        self.weight = nn.Parameter(
            eps * torch.rand(self.nelec, self.nelec)).to(self.device)

    def _backflow_kernel(self, ree):
        """Computes the backflow kernel:

        .. math:
            \\eta(r_{ij}) = w_{ij} r_{ij}^2

        Args:
            r (torch.tensor): e-e distance Nbatch x Nelec x Nelec

        Returns:
            torch.tensor : f(r) Nbatch x Nelec x Nelec
        """
        return self.weight * ree * ree



from .fully_connected_jastrow_kernel import FullyConnectedJastrowKernel
from .jastrow_kernel_electron_electron_nuclei_base import JastrowKernelElectronElectronNucleiBase
from .boys_handy_jastrow_kernel import BoysHandyJastrowKernel



import unittest

import numpy as np
import torch
import torch.optim as optim

from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.utils import (plot_block, plot_blocking_energy,
                            plot_correlation_coefficient,
                            plot_integrated_autocorrelation_time,
                            plot_walkers_traj)
from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterJastrow


__PLOT__ = False


class TestH2Stat(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)

        # optimal parameters
        self.opt_r = 0.69  # the two h are at +0.69 and -0.69
        self.opt_sigma = 1.24

        # molecule
        self.mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # wave function
        self.wf = SlaterJastrow(self.mol, kinetic='jacobi',
                                configs='single(2,2)')

        # sampler
        self.sampler = Metropolis(
            nwalkers=100,
            nstep=500,
            step_size=0.5,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            ntherm=0,
            ndecor=1,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

    def test_sampling_traj(self):

        pos = self.solver.sampler(self.solver.wf.pdf)
        obs = self.solver.sampling_traj(pos)

        plot_walkers_traj(obs.local_energy)
        plot_block(obs.local_energy)

    def test_stat(self):

        pos = self.solver.sampler(self.solver.wf.pdf)
        obs = self.solver.sampling_traj(pos)

        if __PLOT__:
            plot_blocking_energy(obs.local_energy, block_size=10)
            plot_correlation_coefficient(obs.local_energy)
            plot_integrated_autocorrelation_time(obs.local_energy)


if __name__ == "__main__":
    # unittest.main()
    t = TestH2Stat()
    t.setUp()
    t.test_sampling_traj()
    t.test_stat()
    # t.test1_single_point()
    # t.test3_wf_opt()



import numpy as np
import torch
from torch.distributions import MultivariateNormal
from typing import Union, Dict
from .. import log


class Walkers(object):

    def __init__(self, nwalkers: int = 100, nelec: int = 1, ndim: int = 3,
                 init: Union[Dict, None] = None, cuda: bool = False):
        """Creates Walkers for the sampler.

        Args:
            nwalkers (int, optional): Number of walkers. Defaults to 100.
            nelec (int, optional): number of electron. Defaults to 1.
            ndim (int, optional): Number of dimensions. Defaults to 3.
            init (dict, optional): method to initialize the walkers. Defaults to None. (see Molecule.domain())
            cuda (bool, optional): turn cuda ON/OFF. Defaults to False
        """
        self.nwalkers = nwalkers
        self.ndim = ndim
        self.nelec = nelec
        self.init_domain = init

        self.pos = None
        self.status = None

        self.cuda = cuda

        if cuda:
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

    def initialize(self, pos: Union[None, torch.Tensor] = None):
        """Initalize the position of the walkers

        Args:
            method (str, optional): how to initialize the positions. Defaults to 'uniform'.
            pos ([type], optional): existing position of the walkers. Defaults to None.

        Raises:
            ValueError: if the method is not recognized
        """
        if self.cuda:
            self.device = torch.device('cuda')

        if pos is not None:
            if len(pos) > self.nwalkers:
                pos = pos[-self.nwalkers:, :]
            self.pos = pos

        else:
            log.debug("  Initialize walkers")
            if 'center' in self.init_domain.keys():
                self.pos = self._init_center()

            elif 'min' in self.init_domain.keys():
                self.pos = self._init_uniform()

            elif 'mean' in self.init_domain.keys():
                self.pos = self._init_multivar()

            elif 'atom_coords' in self.init_domain.keys():
                self.pos = self._init_atomic()

            else:
                raise ValueError('Init walkers not recognized')

    def _init_center(self):
        """Initialize the walkers at the center of the molecule

        Returns:
            torch.tensor: positions of the walkers
        """
        eps = 1E-3
        pos = -eps + 2 * eps * \
            torch.rand(self.nwalkers, self.nelec * self.ndim)
        return pos.type(
            torch.get_default_dtype()).to(
            device=self.device)

    def _init_uniform(self):
        """Initialize the walkers in a box covering the molecule

        Returns:
            torch.tensor: positions of the walkers
        """
        pos = torch.rand(self.nwalkers, self.nelec * self.ndim)
        pos *= (self.init_domain['max'] - self.init_domain['min'])
        pos += self.init_domain['min']
        return pos.type(
            torch.get_default_dtype()).to(
            device=self.device)

    def _init_multivar(self):
        """Initialize the walkers in a sphere covering the molecule

        Returns:
            torch.tensor -- positions of the walkers
        """
        multi = MultivariateNormal(
            torch.as_tensor(self.init_domain['mean']),
            torch.as_tensor(self.init_domain['sigma']))
        pos = multi.sample((self.nwalkers, self.nelec)).type(
            torch.get_default_dtype())
        pos = pos.view(self.nwalkers, self.nelec * self.ndim)
        return pos.to(device=self.device)

    def _init_atomic(self):
        """Initialize the walkers around the atoms

        Returns:
            torch.tensor -- positions of the walkers
        """
        pos = torch.zeros(self.nwalkers, self.nelec * self.ndim)
        idx_ref, nelec_tot = [], 0

        nelec_placed, natom = [], 0
        for iat, nelec in enumerate(self.init_domain['atom_nelec']):
            idx_ref += [iat] * nelec
            nelec_tot += nelec
            natom += 1

        for iw in range(self.nwalkers):

            nelec_placed = [0] * natom
            idx = torch.as_tensor(idx_ref)
            idx = idx[torch.randperm(nelec_tot)]
            xyz = torch.as_tensor(
                self.init_domain['atom_coords'])[
                idx, :]

            for ielec in range(nelec_tot):
                _idx = idx[ielec]
                if nelec_placed[_idx] == 0:
                    s = 1. / self.init_domain['atom_num'][_idx]
                elif nelec_placed[_idx] < 5:
                    s = 2. / (self.init_domain['atom_num'][_idx] - 2)
                else:
                    s = 3. / (self.init_domain['atom_num'][_idx] - 3)
                xyz[ielec,
                    :] += np.random.normal(scale=s, size=(1, 3))
                nelec_placed[_idx] += 1

            pos[iw, :] = xyz.view(-1)
        return pos.to(device=self.device)



import unittest

import numpy as np
import torch

from qmctorch.wavefunction.orbitals.spherical_harmonics import Harmonics


class TestSphericalHarmonics(unittest.TestCase):

    def setUp(self):
        bas_l = torch.Tensor([0, 1, 1, 1, 2, 2, 2, 2, 2])
        bas_m = torch.Tensor([0, -1, 0, 1, -2, -1, 0, 1, 2])

        self.harmonics = Harmonics('sph', bas_l=bas_l, bas_m=bas_m)
        self.pos = torch.rand(5, 4, 9, 3)

    def test_value(self):
        self.harmonics(self.pos, derivative=0)

    def test_grad(self):
        self.harmonics(self.pos, derivative=1, sum_grad=False)

    def test_jac(self):
        self.harmonics(self.pos, derivative=1, sum_grad=True)

    def test_lap(self):
        self.harmonics(self.pos, derivative=2)


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn
from torch.autograd import grad


class JastrowKernelElectronNucleiBase(nn.Module):

    def __init__(self, nup, ndown, atomic_pos, cuda, **kwargs):
        r"""Base class for the elec-nuc jastrow factor

        .. math::

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            atoms (torch.tensor): atomic positions of the atoms
            w (float, optional): Value of the variational parameter. Defaults to 1..
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__()
        self.nup, self.ndown = nup, ndown
        self.cuda = cuda

        self.nelec = nup + ndown
        self.atoms = atomic_pos
        self.natoms = atomic_pos.shape[0]
        self.ndim = 3

        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')
        self.requires_autograd = True

    def forward(self, r):
        r"""Get the elements of the jastrow matrix :
        .. math::
            out_{i,j} = \exp{ \frac{b r_{i,j}}{1+b'r_{i,j}} }

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the jastrow elements
                          Nbatch x Nelec x Nelec
        """
        raise NotImplementedError()

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons

        .. math::

            d B_{ij} / d k_i =  d B_{ij} / d k_j  = - d B_{ji} / d k_i

            out_{k,i,j} = A1 + A2
            A1_{kij} = w0 \frac{dr_{ij}}{dk_i} / (1 + w r_{ij})
            A2_{kij} = - w0 w' r_{ij} \frac{dr_{ij}}{dk_i} / (1 + w r_{ij})^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """
        if r.requires_grad == False:
            r.requires_grad = True

        with torch.enable_grad():

            kernel = self.forward(r)
            ker_grad = self._grads(kernel, r)

        return ker_grad.unsqueeze(1) * dr

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron

        .. math ::

            d^2 B_{ij} / d k_i^2 =  d^2 B_{ij} / d k_j^2 = d^2 B_{ji} / d k_i^2

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
            d2r (torch.tensor): matrix of the 2nd derivative of
                                the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of
                          the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        dr2 = dr * dr

        if r.requires_grad == False:
            r.requires_grad = True

        with torch.enable_grad():

            kernel = self.forward(r)

            ker_hess, ker_grad = self._hess(kernel, r)

            jhess = (ker_hess).unsqueeze(1) * \
                dr2 + ker_grad.unsqueeze(1) * d2r

        return jhess

    @staticmethod
    def _grads(val, pos):
        """Get the gradients of the jastrow values
        of a given orbital terms

        Args:
            pos ([type]): [description]

        Returns:
            [type]: [description]
        """
        return grad(val, pos, grad_outputs=torch.ones_like(val))[0]

    @staticmethod
    def _hess(val, pos):
        """get the hessian of the jastrow values.
        of a given orbital terms
        Warning thos work only because the orbital term are dependent
        of a single rij term, i.e. fij = f(rij)

        Args:
            pos ([type]): [description]
        """

        gval = grad(val,
                    pos,
                    grad_outputs=torch.ones_like(val),
                    create_graph=True)[0]

        hval = grad(gval, pos, grad_outputs=torch.ones_like(gval))[0]

        return hval, gval



import torch
from torch import nn
from .scaling import (get_scaled_distance,
                      get_der_scaled_distance,
                      get_second_der_scaled_distance)


class ElectronElectronDistance(nn.Module):

    def __init__(self, nelec, ndim=3, scale=False, scale_factor=0.6):
        """Computes the electron-electron distances

        .. math::
            r_{ij} = \\sqrt{ (x_i-x_j)^2 + (y_i-y_j)^2 + (z_i-z_j)^2}

        Args:
            nelec (int): number of electrons
            ndim (int): number of spatial dimensions
            scale(bool, optional): return scaled values, Defaults to False
            scale_factor(float, optional): value of the scale factor,
                                           Defaults to 0.6

        Examples::
            >>> edist = ElectronDistance(2,3)
            >>> pos = torch.tensor(500,6)
            >>> r = edist(pos)
            >>> dr = edist(pos,derivative=1)

        """

        super().__init__()
        self.nelec = nelec
        self.ndim = ndim
        self.scale = scale
        self.kappa = scale_factor

        _type_ = torch.get_default_dtype()
        if _type_ == torch.float32:
            self.eps = 1E-6
        elif _type_ == torch.float64:
            self.eps = 1E-16

    def forward(self, input, derivative=0):
        """Compute the pairwise distance between the electrons
        or its derivative. \n

        When required, the derivative is computed wrt to the first electron i.e.

        .. math::
            \\frac{dr_{ij}}{dx_i}

        which is different from :

        .. math::
            \\frac{d r_{ij}}{dx_j} = -\\frac{dr_{ij}}{dx_i}

        Args:
            input (torch.tesnor): position of the electron \n
                                  size : Nbatch x [Nelec x Ndim]
            derivative (int, optional): degre of the derivative. \n
                                        Defaults to 0.

        Returns:
            torch.tensor: distance (or derivative) matrix \n
                          Nbatch x Nelec x Nelec if derivative = 0 \n
                          Nbatch x Ndim x  Nelec x Nelec if derivative = 1,2

        """

        # get the distance matrices
        input_ = input.view(-1, self.nelec, self.ndim)
        dist = self.get_distance_quadratic(input_)
        dist = self.safe_sqrt(dist)

        if derivative == 0:
            if self.scale:
                return get_scaled_distance(self.kappa, dist)
            else:
                return dist

        elif derivative == 1:

            der_dist = self.get_der_distance(input_, dist)

            if self.scale:
                return get_der_scaled_distance(self.kappa, dist, der_dist)
            else:
                return der_dist

        elif derivative == 2:

            d2_dist = self.get_second_der_distance(input_, dist)

            if self.scale:
                der_dist = self.get_der_distance(input_, dist)
                return get_second_der_scaled_distance(self.kappa,
                                                      dist,
                                                      der_dist,
                                                      d2_dist)
            else:
                return d2_dist

    def safe_sqrt(self, dist):
        """Compute the square root of the electron electron distance matrix.

        Args:
            dist (torch.tensor): ee distances squared
                                 Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: sqrt of dist Nbatch x Nelec x Nelec

        """

        # epsilon on the diag needed for back prop
        eps_ = self.eps * \
            torch.diag(dist.new_ones(dist.shape[-1])).expand_as(dist)

        # extact the diagonal as diag can be negative someties
        # due to numerical noise
        diag = torch.diag_embed(
            torch.diagonal(
                dist, dim1=-1, dim2=-2))

        # remove diagonal and add eps for backprop
        dist = torch.sqrt(dist - diag + eps_)

        return dist

    def get_der_distance(self, pos, dist):
        """Get the derivative of the electron electron distance matrix.

        .. math::
            \\frac{d r_{ij}}{d x_i}

        Args:
            pos (torch.tensor): positions of the electrons
                                Nbatch x Nelec x Ndim
            dist (torch.tensor): distance matrix between the elecs
                           Nbatch x Nelec x Nelec

        Returns:
            [type]: [description]
        """

        eps_ = self.eps * \
            torch.diag(dist.new_ones(
                dist.shape[-1])).expand_as(dist)

        invr = (1. / (dist + eps_)).unsqueeze(1)
        diff_axis = pos.transpose(1, 2).unsqueeze(3)
        diff_axis = diff_axis - diff_axis.transpose(2, 3)
        return diff_axis * invr

    def get_second_der_distance(self, pos, dist):
        """Get the second derivative of the electron electron distance matrix.

        .. math::
            \\frac{d^2 r_{ij}}{d x_i^2}

        Args:
            pos (torch.tensor): positions of the electrons
                                Nbatch x Nelec x Ndim
            dist (torch.tensor): distance matrix between the elecs
                           Nbatch x Nelec x Nelec

        Returns:
            [type]: [description]
        """

        eps_ = self.eps * \
            torch.diag(dist.new_ones(
                dist.shape[-1])).expand_as(dist)
        invr3 = (1. / (dist**3 + eps_)).unsqueeze(1)
        diff_axis = pos.transpose(1, 2).unsqueeze(3)
        diff_axis = (diff_axis - diff_axis.transpose(2, 3))**2

        diff_axis = diff_axis[:, [
            [1, 2], [2, 0], [0, 1]], ...].sum(2)
        return (diff_axis * invr3)

    @staticmethod
    def get_distance_quadratic(pos):
        """Compute the distance following a quadratic expansion

        Arguments:
            pos {torch.tensor} -- electron position [nbatch x nelec x ndim]

        Returns:
            torch.tensor -- distance matrices nbatch x nelec x ndim]
        """

        norm = (pos**2).sum(-1).unsqueeze(-1)
        dist = (norm + norm.transpose(1, 2) - 2.0 *
                torch.bmm(pos, pos.transpose(1, 2)))
        return dist

    @staticmethod
    def get_difference(pos):
        """Compute the difference ri - rj

        Arguments:
            pos {torch.tensor} -- electron position [nbatch x nelec x ndim]

        Returns:
            torch.tensor -- distance matrices nbatch x nelec x nelec x ndim]
        """
        out = pos[:, :, None, :] - pos[:, None, :, :]
        return out



import torch
from torch import nn
from torch.autograd import grad
from torch import nn


class JastrowKernelElectronElectronBase(nn.Module):

    def __init__(self, nup, ndown, cuda, **kwargs):
        r"""Base class for the elec-elec jastrow kernels

        Args:
            nup ([type]): [description]
            down ([type]): [description]
            cuda (bool, optional): [description]. Defaults to False.
        """

        super().__init__()
        self.nup, self.ndown = nup, ndown
        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

        self.requires_autograd = True

    def forward(self, r):
        r"""Get the elements of the jastrow matrix :


        Args:
            r (torch.tensor): matrix of the e-e distances
                            Nbatch x Nelec_pair

        Returns:
            torch.tensor: matrix fof the jastrow elements
                        Nmo x Nbatch x Nelec_pair

        Note:
            The kernel receives a [Nbatch x Npair] tensor.
            The kernel must first reshape that tensor to a [Nbatch*Npair, 1].
            The kernel must process this tensor to another [Nbatch*Npair, 1] tensor.
            The kenrel must reshape the output to a [Nbatch x Npair] tensor.

        Example:
            >>> def forward(self, x):
            >>>     nbatch, npairs = x.shape
            >>>     x = x.reshape(-1, 1)
            >>>     x = self.fc1(x)
            >>>     ...
            >>>     return(x.reshape(nbatch, npairs))
        """
        raise NotImplementedError()

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons using automatic differentiation

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec_pair

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nmo x Nbatch x Ndim x  Nelec_pair
        """

        if r.requires_grad == False:
            r.requires_grad = True

        with torch.enable_grad():

            kernel = self.forward(r)
            ker_grad = self._grads(kernel, r)

        return ker_grad.unsqueeze(1) * dr

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron using automatic differentiation

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec_pair
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec_pair
            d2r (torch.tensor): matrix of the 2nd derivative of
                                the e-e distances
                              Nbatch x Ndim x Nelec_pair

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of
                          the jastrow elements
                          Nmo x Nbatch x Ndim x Nelec_pair
        """
        dr2 = dr * dr

        if r.requires_grad == False:
            r.requires_grad = True

        with torch.enable_grad():

            kernel = self.forward(r)
            ker_hess, ker_grad = self._hess(kernel, r)

            jhess = (ker_hess).unsqueeze(1) * \
                dr2 + ker_grad.unsqueeze(1) * d2r

        return jhess

    @staticmethod
    def _grads(val, pos):
        """Get the gradients of the jastrow values
        of a given orbital terms

        Args:
            pos ([type]): [description]

        Returns:
            [type]: [description]
        """
        return grad(val, pos, grad_outputs=torch.ones_like(val))[0]

    @staticmethod
    def _hess(val, pos):
        """get the hessian of the jastrow values.
        of a given orbital terms
        Warning thos work only because the orbital term are dependent
        of a single rij term, i.e. fij = f(rij)

        Args:
            pos ([type]): [description]
        """

        gval = grad(val,
                    pos,
                    grad_outputs=torch.ones_like(val),
                    create_graph=True)[0]

        hval = grad(gval, pos, grad_outputs=torch.ones_like(gval))[0]

        return hval, gval



from tqdm import tqdm
import torch
from torch.distributions import MultivariateNormal
from time import time
from typing import Callable, Union, Dict
from .sampler_base import SamplerBase
from .. import log


class Metropolis(SamplerBase):

    def __init__(self,
                 nwalkers: int = 100,
                 nstep: int = 1000,
                 step_size: float = 0.2,
                 ntherm: int = -1,
                 ndecor: int = 1,
                 nelec: int = 1,
                 ndim: int = 3,
                 init: Dict = {'min': -5, 'max': 5},
                 move: Dict = {'type': 'all-elec', 'proba': 'normal'},
                 cuda: bool = False):
        """Metropolis Hasting generator

        Args:
            nwalkers (int, optional): Number of walkers. Defaults to 100.
            nstep (int, optional): Number of steps. Defaults to 1000.
            step_size (int, optional): length of the step. Defaults to 0.2.
            nelec (int, optional): total number of electrons. Defaults to 1.
            ntherm (int, optional): number of mc step to thermalize. Defaults to -1, i.e. keep only the last position
            ndecor (int, optional): number of mc step for decorelation. Defauts to 1.
            ndim (int, optional): total number of dimension. Defaults to 3.
            init (dict, optional): method to init the positions of the walkers. See Molecule.domain()

            move (dict, optional): method to move the electrons. default('all-elec','normal') \n
                                   'type':
                                        'one-elec': move a single electron per iteration \n
                                        'all-elec': move all electrons at the same time \n
                                        'all-elec-iter': move all electrons by iterating through single elec moves \n
                                    'proba' :
                                        'uniform': uniform in a cube \n
                                        'normal': gussian in a sphere \n
            cuda (bool, optional): turn CUDA ON/OFF. Defaults to False.


        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> wf = SlaterJastrow(mol)
            >>> sampler = Metropolis(nwalkers=100, nelec=wf.nelec)
            >>> pos = sampler(wf.pdf)
        """

        SamplerBase.__init__(self, nwalkers, nstep,
                             step_size, ntherm, ndecor,
                             nelec, ndim, init, cuda)

        self.configure_move(move)
        self.log_data()

    def log_data(self):
        """log data about the sampler."""
        log.info('  Move type           : {0}', self.movedict['type'])
        log.info(
            '  Move proba          : {0}', self.movedict['proba'])

    def __call__(self, pdf: Callable, pos: Union[None, torch.Tensor] = None,
                 with_tqdm: bool = True) -> torch.Tensor:
        """Generate a series of point using MC sampling

        Args:
            pdf (callable): probability distribution function to be sampled
            pos (torch.tensor, optional): position to start with.
                                          Defaults to None.
            with_tqdm (bool, optional): use tqdm progress bar. Defaults to True.

        Returns:
            torch.tensor: positions of the walkers
        """

        _type_ = torch.get_default_dtype()
        if _type_ == torch.float32:
            eps = 1E-7
        elif _type_ == torch.float64:
            eps = 1E-16

        if self.ntherm >= self.nstep:
            raise ValueError('Thermalisation longer than trajectory')

        with torch.no_grad():

            if self.ntherm < 0:
                self.ntherm = self.nstep + self.ntherm

            self.walkers.initialize(pos=pos)
            fx = pdf(self.walkers.pos)

            fx[fx == 0] = eps
            pos, rate, idecor = [], 0, 0

            rng = tqdm(range(self.nstep),
                       desc='INFO:QMCTorch|  Sampling',
                       disable=not with_tqdm)
            tstart = time()

            for istep in rng:

                for id_elec in self.fixed_id_elec_list:

                    # new positions
                    Xn = self.move(pdf, id_elec)

                    # new function
                    fxn = pdf(Xn)
                    fxn[fxn == 0.] = eps
                    df = fxn / fx

                    # accept the moves
                    index = self._accept(df)

                    # acceptance rate
                    rate += index.byte().sum().float().to('cpu') / \
                        (self.nwalkers * self._move_per_iter)

                    # update position/function value
                    self.walkers.pos[index, :] = Xn[index, :]
                    fx[index] = fxn[index]
                    fx[fx == 0] = eps

                if (istep >= self.ntherm):
                    if (idecor % self.ndecor == 0):
                        pos.append(self.walkers.pos.to('cpu').clone())
                    idecor += 1

            if with_tqdm:
                log.info(
                    "   Acceptance rate     : {:1.2f} %", (rate / self.nstep * 100))
                log.info(
                    "   Timing statistics   : {:1.2f} steps/sec.", self.nstep/(time()-tstart))
                log.info(
                    "   Total Time          : {:1.2f} sec.", (time()-tstart))

        return torch.cat(pos).requires_grad_()

    def configure_move(self, move: Dict):
        """Configure the electron moves

        Args:
            move (dict, optional): method to move the electrons. default('all-elec','normal') \n
                                   'type':
                                        'one-elec': move a single electron per iteration \n
                                        'all-elec': move all electrons at the same time \n
                                        'all-elec-iter': move all electrons by iterating through single elec moves \n
                                    'proba' :
                                        'uniform': uniform ina cube \n
                                        'normal': gussian in a sphere \n

        Raises:
            ValueError: If moves are not recognized
        """

        self.movedict = move

        if 'type' not in self.movedict.keys():
            print('Metroplis : Set 1 electron move by default')
            self.movedict['type'] = 'one-elec'

        if 'proba' not in self.movedict.keys():
            print('Metroplis : Set uniform trial move probability')
            self.movedict['proba'] = 'uniform'

        if self.movedict['proba'] == 'normal':
            _sigma = self.step_size / \
                (2 * torch.sqrt(2 * torch.log(torch.as_tensor(2.))))
            self.multiVariate = MultivariateNormal(
                torch.zeros(self.ndim), _sigma * torch.eye(self.ndim))

        self._move_per_iter = 1
        if self.movedict['type'] not in [
                'one-elec', 'all-elec', 'all-elec-iter']:
            raise ValueError(
                " 'type' in move should be 'one-elec','all-elec', \
                  'all-elec-iter'")

        if self.movedict['type'] == 'all-elec-iter':
            self.fixed_id_elec_list = range(self.nelec)
            self._move_per_iter = self.nelec
        else:
            self.fixed_id_elec_list = [None]

    def move(self, pdf: Callable, id_elec: int) -> torch.Tensor:
        """Move electron one at a time in a vectorized way.

        Args:
            pdf (callable): function to sample
            id_elec (int): index f the electron to move

        Returns:
            torch.tensor: new positions of the walkers
        """
        if self.nelec == 1 or self.movedict['type'] == 'all-elec':
            return self.walkers.pos + self._move(self.nelec)

        else:

            # clone and reshape data : Nwlaker, Nelec, Ndim
            new_pos = self.walkers.pos.clone()
            new_pos = new_pos.view(self.nwalkers,
                                   self.nelec, self.ndim)

            # get indexes
            if id_elec is None:
                index = torch.LongTensor(self.nwalkers).random_(
                    0, self.nelec)
            else:
                index = torch.LongTensor(self.nwalkers).fill_(id_elec)

            # change selected data
            new_pos[range(self.nwalkers), index,
                    :] += self._move(1)

            return new_pos.view(self.nwalkers, self.nelec * self.ndim)

    def _move(self, num_elec: int) -> torch.Tensor:
        """propose a move for the electrons

        Args:
            num_elec (int): number of electrons to move

        Returns:
            torch.tensor: new positions of the walkers
        """
        if self.movedict['proba'] == 'uniform':
            d = torch.rand(
                (self.nwalkers, num_elec, self.ndim), device=self.device).view(
                self.nwalkers, num_elec * self.ndim)
            return self.step_size * (2. * d - 1.)

        elif self.movedict['proba'] == 'normal':
            displacement = self.multiVariate.sample(
                (self.nwalkers, num_elec)).to(self.device)
            return displacement.view(
                self.nwalkers, num_elec * self.ndim)

    def _accept(self, proba: torch.Tensor) -> torch.Tensor:
        """accept the move or not

        Args:
            proba (torch.tensor): probability of each move

        Returns:
            t0rch.tensor: the indx of the accepted moves
        """

        proba[proba > 1] = 1.0
        tau = torch.rand_like(proba)
        index = (proba - tau >= 0).reshape(-1)
        return index.type(torch.bool)
